import pytest

import math
import numpy as np
import warnings

from cosapp.core.eval_str import EvalString
from cosapp.systems import System
from typing import Dict, Set


def test_EvalString__init__(eval_context):
    e = EvalString('3.14159', eval_context)
    assert e.eval_context is eval_context
    assert e.constant
    assert len(e.locals) == 0

    e = EvalString('22 + a', eval_context)
    assert e.eval_context is eval_context
    assert not e.constant
    assert set(e.locals) == {'a'}
    assert e.locals['a'] == eval_context.a


def test_EvalString_functions(ufunc_test_data):
    """Check that all tested symbols are actually implemented"""
    tested = set(ufunc_test_data)
    available = set(EvalString.available_symbols())
    available.remove('__builtins__')
    assert tested <= available


@pytest.mark.parametrize("fname", [
        func_name for func_name in EvalString.available_symbols()
        if not func_name.startswith('u')
])
def test_EvalString_exec_comp_value(fname, ufunc_test_data):
    class Test(System):
        def setup(self, **kwargs):
            for name, value in kwargs.items():
                self.add_inward(name, value=value)

    try:
        test_data = ufunc_test_data[fname]
    except KeyError:
        pytest.fail(reason=f"function {fname!r} is not tested")

    try:
        func = test_data['func']
    except KeyError:
        expected = test_data['check_val']
    else:
        expected = func(*test_data['args'].values())

    with warnings.catch_warnings():
        warnings.simplefilter('error')  # 'default', 'ignore', or 'error'
        s = Test('tmp', **test_data.get('args', {}))
        eval_str = EvalString(test_data['expr'], s)

    assert eval_str.eval() == pytest.approx(expected, rel=1e-14)


@pytest.mark.skip(reason='TODO')
def test_EvalString_residue_as_context(self):
    pass


@pytest.mark.parametrize("expression, expected", [
    (1, 1),
    (-1, -1),
    (0.2, 0.2),
    ("--1", 1),
    ("0.2", 0.2),
    ("0.2 / 0.5", pytest.approx(0.4, rel=1e-14)),
    ("cos(pi)", pytest.approx(-1, rel=1e-14)),
    ("sin(pi / 2)", pytest.approx(1, rel=1e-14)),
    ("log(exp(2))", pytest.approx(2, rel=1e-14)),
    ("log(e)", pytest.approx(1, rel=1e-14)),
    (None, None),
    ("None", None),
    (dict(), dict()),
    ("[]", list()),
    (True, True),
    (False, False),
    ("True", True),
    ("False", False),
    ("0.3 < 0.7", True),
    ("0.3 <= 0.3", True),
    ("0.3 < 0.7 < 0.6", False),
    ("0.3 + 0.7 == 1", True),
    ("[0., 0., 0.]", pytest.approx([0., 0., 0.], abs=0)),
    ("[0.] * 3", pytest.approx([0., 0., 0.], abs=0)),
    ("array([0, 0, 0], dtype=float)", pytest.approx([0., 0., 0.], abs=0)),
    ("ones(4, dtype=int)", pytest.approx([1, 1, 1, 1], abs=0)),
    ("ones(2, dtype=bool)", pytest.approx([True, True], abs=0)),
    (repr(np.zeros(3)), pytest.approx([0., 0., 0.], abs=0)),
    ("zeros(3)", pytest.approx([0., 0., 0.], abs=0)),
    ("ones(3)", pytest.approx([1., 1., 1.], abs=0)),
    (np.zeros(3), pytest.approx([0., 0., 0.], abs=0)),
    (np.zeros((2, 5, 3)), pytest.approx(np.zeros((2, 5, 3)), abs=0)),
    (np.ones(3), pytest.approx([1., 1., 1.], abs=0)),
    ("-e", pytest.approx(-math.e, rel=1e-14)),
    ("9 + 3 + 6", 18),
    ("9 + 3 / 11", pytest.approx(9 + 3 / 11, rel=1e-14)),
    ("(9 + 3) / 11", pytest.approx(12 / 11, rel=1e-14)),
    ("(9 + 3)", 12),
    ("9 - 12 - 6", -9),
    ("9 - (12 - 6)", 9 - (12 - 6)),
    ("2 * 3.14159",  pytest.approx(2 * 3.14159, rel=1e-14)),
    ("3.1415926535 * 3.1415926535 / 10", pytest.approx(3.1415926535 * 3.1415926535 / 10, rel=1e-14)),
    ("pi * pi / 10", pytest.approx(math.pi * math.pi / 10, rel=1e-14)),
    ("pi**2 / 10", pytest.approx(math.pi**2 / 10, rel=1e-14)),
    ("pi**2", pytest.approx(math.pi ** 2, rel=1e-14)),
    ("round(pi**2)", pytest.approx(round(math.pi ** 2), rel=1e-14)),
    ("6.02E23 * 8.048", pytest.approx(6.02e23 * 8.048, rel=1e-14)),
    ("e / 3", pytest.approx(math.e / 3, rel=1e-14)),
    ("round(e)", pytest.approx(round(math.e), rel=1e-14)),
    ("round(-e)", pytest.approx(round(-math.e), rel=1e-14)),
    ("e**pi", pytest.approx(math.e ** math.pi, rel=1e-14)),
    ("2**3**2", pytest.approx(2 ** 3 ** 2, rel=1e-14)),
    ("2**3+2", pytest.approx(2 ** 3 + 2, rel=1e-14)),
    ("2**9", pytest.approx(2 ** 9, rel=1e-14)),
    ("{1, 2, 3, 3, 2, }", {1, 2, 3}),
    ("acos(-1)", pytest.approx(np.pi, rel=1e-14)),
    ("arccos(-1)", pytest.approx(np.pi, rel=1e-14)),
])
def test_EvalString_constant_expr(eval_context, expression, expected):
    """Test expressions expected to be interpreted as constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert s.constant
    assert s.eval() == expected


@pytest.mark.parametrize("expression, expected", [
    ("-2 * g", dict(value=pytest.approx(-2 * 9.80665, rel=1e-14), constant=True)),
    ("1e-23 * NA + g", dict(value=pytest.approx(15.828790), constant=True)),
    ("g * x", dict(constant=False)),
])
def test_EvalString_with_constants(eval_context, expression, expected):
    """Test expressions with system constants, expected to be interpreted as constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert s.constant == expected.get('constant', False)
    if 'value' in expected:
        assert s.eval() == expected['value']


def test_EvalString_subsystem_constant():
    """Test expressions involving sub-system constants"""
    class SystemWithConstants(System):
        def setup(self, constants: dict={}):
            for name, value in constants.items():
                self.add_property(name, value)

    top = System('top')
    mid = top.add_child(SystemWithConstants('mid', constants={'g': 9.81}))
    sub = mid.add_child(SystemWithConstants('sub', constants={'c': 0.10}))

    s = EvalString('2 * c', sub)
    assert s.variables() == set()
    assert s.variables(include_const=True) == {'c'}
    assert s.constant
    assert s.eval() == pytest.approx(0.2, rel=1e-14)

    s = EvalString('2 * sub.c', mid)
    assert s.variables() == set()
    assert s.variables(include_const=True) == {'sub.c'}
    assert s.constant
    assert s.eval() == pytest.approx(0.2, rel=1e-14)

    s = EvalString('2 * mid.sub.c', top)
    assert s.variables() == set()
    assert s.variables(include_const=True) == {'mid.sub.c'}
    assert s.constant
    assert s.eval() == pytest.approx(0.2, rel=1e-14)

    s = EvalString('mid.g * mid.sub.c', top)
    assert s.variables() == set()
    assert s.variables(include_const=True) == {'mid.g', 'mid.sub.c'}
    assert s.constant
    assert s.eval() == pytest.approx(0.981, rel=1e-14)


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", 3.14),
    ("0.2 * a / 0.5", pytest.approx(0.8, rel=1e-14)),
    ("log10(x[0])", pytest.approx(-1, rel=1e-14)),
    ("a + b - x[1]", pytest.approx(2.7, rel=1e-14)),
    ("a - a + 1", 1),
    ("9 + sub.in_.q / 10", pytest.approx(9.5, rel=1e-14)),
    ("9 + sin(sub.in_.q / 11)", pytest.approx(9 + math.sin(5 / 11), rel=1e-14)),
    ("out.q - sub.in_.q", pytest.approx(-4.5, rel=1e-14)),
    ("out. q - sub.in_ .  q", pytest.approx(-4.5, rel=1e-14)),
    ("concatenate((x, [-a, sub.in_.q]))", pytest.approx([0.1, -0.2, -3.14, -2, 5], rel=1e-14)),
    ("len(out)", 1),   # not constant, by convention
])
def test_EvalString_nonconstant_expr(eval_context, expression, expected):
    """Test expressions expected to be interpreted as non-constant"""
    s = EvalString(expression, eval_context)
    assert s.eval_context is eval_context
    assert not s.constant
    assert s.eval() == expected


@pytest.mark.parametrize("expression, exception", [
    ("a = 2.0", SyntaxError),
    ("cos(pi", SyntaxError),
    ("a + .  5", SyntaxError),
    ("1 + _", NameError),
    ("1 / 0", ZeroDivisionError),
    ("foo.bar(1)", NameError),
    ("a + ['v', True]", TypeError),
    ("out.monkey", AttributeError),
    ("sub.out.q", AttributeError),
    ("sub.in_.", SyntaxError),
    ("sub.in_..", SyntaxError),
    ("", ValueError),
    ("  ", ValueError),
])
def test_EvalString_erroneous_expr(eval_context, expression, exception):
    """Test erroneous expressions expected to raise an exception at instantiation"""
    with pytest.raises(exception):
        EvalString(expression, eval_context)


@pytest.mark.parametrize("expression, expected", [
    ("1", "1"),
    ("1.2", "1.2"),
    (8, "8"),
    (1.3, "1.3"),
    ([0.1, 0.2, 0.3], "[0.1, 0.2, 0.3]"),
    (np.r_[0.1, 0.2, 0.3], "array([0.1, 0.2, 0.3])"),
    ("g * cos(pi * x) + sub.z", "g * cos(pi * x) + sub.z"),
    ("out. q - sub.in_   .  q  -  sub   .z * .5", "out.q - sub.in_.q  -  sub.z * .5"),
])
def test_EvalString_str_repr(eval_context, expression, expected):
    s = EvalString(expression, eval_context)
    assert str(s) == expected
    assert repr(s) == repr(expected)  # same repr as built-in Python strings


@pytest.mark.parametrize("expression, patterns, results", [
    ("norm(x, inf)", ["norm", "inf", "x", "foo"], [True] * 3 + [False]),
    ("0.2 * a / 0.5", ["0.", "a", "x"], [True, True, False]),
    ("x", ["x", "Y", "foo"], [True, False, False]),
    ("[a, 2]", ["a", "2", "1", "foo"], [True, True, False, False]),
    ("[1, 2]", ["a", "2", "1", "1, 2"], [False, True, True, True]),
])
def test_EvalString_contains(eval_context, expression, patterns, results):
    """Test expressions expected to be interpreted as non-constant"""
    s = EvalString(expression, eval_context)
    for pattern, expected in zip(patterns, results):
        assert (pattern in s) == expected


@pytest.mark.parametrize("var", [
    "a", "b", "out.q", "sub.z", "sub.in_.q", ])
def test_EvalString_multiple_eval(eval_context, var):
    s = EvalString(var, eval_context)
    assert s.eval() == eval_context[var]
    for eval_context[var] in [0, 0.1, -22]:
        assert s.eval() == eval_context[var]


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", "norm(x, inf)"),
    ("123.456", "123.456"),
    (123.456, "123.456"),
    ([1, 2, 3, ], "[1, 2, 3]"),
    ((1, 2, 3, ), "(1, 2, 3)"),
    ({1, 2, 3}, "{1, 2, 3}"),
    ("{1, 2, 3, 3, 2, }", "{1, 2, 3, 3, 2, }"),
    (np.r_[0.1, 0.2, 0.3], "array([0.1, 0.2, 0.3])"),
    (np.array([0, 0, 0], dtype=float), "array([0., 0., 0.])"),
    (np.ones(2, dtype=int), "array([1, 1])"),
    (None, "None"),
    ("out. q - sub.in_   .  q  -  sub   .z * .5", "out.q - sub.in_.q  -  sub.z * .5"),
])
def test_EvalString_string(eval_context, expression, expected):
    assert EvalString.string(expression) == expected
    s = EvalString(expression, eval_context)
    assert EvalString.string(s) == expected


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", {'x'}),
    ("g * cos(pi * x)", {'x', 'g'}),
    ("g * cos(pi * x) + sub.z", {'x', 'g', 'sub'}),
    ("g * cos(pi * x) + sub.z + out.q", {'x', 'g', 'sub', 'out'}),
    ("g * cos(pi * x) / out.q + sub.z * sub.in_.q + out.q", {'x', 'g', 'sub', 'out'}),
])
def test_EvalString_locals(eval_context, expression, expected):
    e = EvalString(expression, eval_context)
    assert set(e.locals) == expected


@pytest.mark.parametrize("expression, expected", [
    ("norm(x, inf)", dict(variables={'x'})),
    ("g * cos(pi * x)", dict(variables={'x'}, constants={'g'})),
    ("g * cos(pi * x) + sub.z", dict(variables={'x', 'sub.z'}, constants={'g'})),
    (
        "g * cos(pi * x) / out.q + sub.z * sub.in_.q + out.q",
        dict(variables={'x', 'sub.z', 'sub.in_.q', 'out.q'}, constants={'g'})
    ),
    (
        "(g * cos(pi * x), out.q, sub.z, sub.in_.q, 2 * out.q)",
        dict(variables={'x', 'sub.z', 'sub.in_.q', 'out.q'}, constants={'g'}),
    ),
    ("out.q + sub.z * sub.in_.q + B52.in_.q", dict(variables={'sub.z', 'sub.in_.q', 'out.q', 'B52.in_.q'})),
    ("out. q + sub  .z * sub . in_.q + out  . q", dict(variables={'sub.z', 'sub.in_.q', 'out.q'})),
    ("len(out)", dict(variables=set())),
])
def test_EvalString_variables(eval_context, expression, expected: Dict[str, Set[str]]):
    """Test method `EvalString.variables()`.
    """
    expected.setdefault('constants', set())
    all_required = expected['variables'].union(expected['constants'])
    
    e = EvalString(expression, eval_context)
    assert e.variables() == expected['variables']
    assert e.variables(include_const=True) == all_required



"""Utility functions for testing purposes"""
import numpy
import pytest
import inspect
import warnings
import itertools
from numbers import Number
from contextlib import contextmanager
from typing import Tuple, Dict, Any, Union, Iterable, Type, Optional
from cosapp.base import System


ArgsKwargs = Tuple[Tuple[Any], Dict[str, Any]]


def assert_keys(dictionary, *keys):
    """Utility function to test dictionary keys"""
    if set(dictionary.keys()) != set(keys):
        actual = list(dictionary.keys())
        raise KeyError(f"actual: {actual}; expected: {list(keys)}") 


def assert_all_type(collection, dtype):
    """Asserts that all the elements of a collecion are of a given type"""
    def assert_type(name, value):
        if not isinstance(value, dtype):
            actual = type(value).__name__
            raise TypeError(
                f"{name} is of type {actual}; expected {dtype.__name__}"
            ) 
    if isinstance(collection, dict):
        for key, value in collection.items():
            assert_type(f"element with key '{key}'", value)
    else:
        for i, value in enumerate(collection):
            assert_type(f"element #{i}", value)


def rel_error(actual: Union[Number, Iterable], expected: Union[Number, Iterable]) -> Union[float, numpy.ndarray]:
    """Computes the relative error of `actual` compared to `expected`
    """
    error = lambda a, x: abs(a) if x == 0 else abs(a / x - 1)

    if isinstance(actual, Number):
        return error(actual, expected)
    
    actual = numpy.asarray(actual)

    if isinstance(expected, Number):
        iterator = (error(a, expected) for a in actual.flat)
    else:
        iterator = itertools.starmap(error,
            zip(actual.flat, numpy.ravel(expected))
        )
    errors = numpy.fromiter(
        iterator,
        dtype=float,
        count=actual.size,
    )
    return errors.reshape(actual.shape)


def assert_close_dict(actual: dict, expected: dict, abs=None, rel=None) -> None:
    """Assert that `actual` and `expected` dictionaries are identical, within given tolerance bounds.
    Works recursively with nested dictionaries.
    """
    def assert_close(actual: dict, expected: dict, context=""):
        """Recursive test function, with contextual message"""
        assert set(actual) == set(expected), context
        for key, value in expected.items():
            local_context = f"{context}[{key!r}]"
            if isinstance(value, dict):
                assert_close(actual[key], value, context=local_context)
            else:
                try:
                    expected_value = pytest.approx(value, abs=abs, rel=rel)
                except TypeError:
                    expected_value = value
                assert actual[key] == expected_value, f"key {local_context}"

    assert_close(actual, expected)


def get_args(*args, **kwargs) -> ArgsKwargs:
    """Utility function to collect args and kwargs in a tuple"""
    return args, kwargs


@contextmanager
def no_exception():
    """Context manager to assert that a block does not raise any exception"""
    try:
        yield

    except Exception as error:
        raise AssertionError(f"Unexpected exception raised: {error!r}")


@contextmanager
def not_raised(ExpectedException):
    """Context manager to assert that a block does not raise `ExpectedException`"""
    # https://gist.github.com/oisinmulvihill/45c14271fad7794a4a52516ecb784e69
    try:
        yield

    except ExpectedException as error:
        raise AssertionError(f"Raised {error!r} exception when it should not!")

    except Exception as error:
        raise AssertionError(f"Unexpected exception raised: {error!r}")


def DummySystemFactory(
    classname: str,
    base: Optional[Type[System]]=None,
    **settings
) -> Type[System]:
    """Factory creating a dummy system class with custom attributes.
    System is "dummy" in the sense it has no compute, and no connectors.
    
    Parameters:
    -----------
    - classname [str]: Output class name
    - base [type[System], optional]:
        Base class, derived from `System`. If not provided (default), base is `System`.
    - **settings [dict[str, args_kwargs]]:
        Class characteristics, as a dictionary. Keys are attribute names (e.g. inputs);
        values are (args, kwargs) forwarded to the associated method (e.g. `add_input`).
    
    Possible Attributes:
    --------------------
    - inputs
    - outputs
    - inwards
    - outwards
    - modevars_in
    - modevars_out
    - transients
    - rates
    - properties
    - children
    - events
    - unknowns
    - equations
    - targets
    - design_methods

    Examples:
    ---------
    >>> from cosapp.utils.testing import DummySystemFactory, get_args
    >>> 
    >>> Dummy = DummySystemFactory(
    >>>     inwards=[
    >>>         get_args('h', 0.1, unit='m'),
    >>>         get_args('L', 2.0, unit='m'),
    >>>     ],
    >>>     outwards=[
    >>>         get_args('b_ratio', 0.0),
    >>>     ],
    >>>     events=[
    >>>         get_args('kaboom', trigger='h > L / 2')
    >>>     ],
    >>>     properties=[
    >>>         get_args('n', 12),
    >>>     ],
    >>>     equations=[
    >>>         "b_ratio == 1",
    >>>     ],
    >>>     unknowns=[
    >>>         "h",
    >>>     ],
    >>> )
    >>> 
    >>> s = Dummy('s')
    >>> assert s.assembled_problem().shape == (1, 1)
    """
    # mapping option / method
    # for example: `inputs` <-> `add_input`
    struct_method_mapping = {
        "inputs": "add_input",
        "outputs": "add_output",
        "inwards": "add_inward",
        "outwards": "add_outward",
        "modevars_in": "add_inward_modevar",
        "modevars_out": "add_outward_modevar",
        "transients": "add_transient",
        "rates": "add_rate",
        "properties": "add_property",
        "children": "add_child",
    }
    extra_method_mapping = {
        "events": "add_event",
        "unknowns": "add_unknown",
        "equations": "add_equation",
        "targets": "add_target",
        "design_methods": "add_design_method",
    }
    unknown_keys = set(settings).difference(
        struct_method_mapping,
        extra_method_mapping,
    )
    if unknown_keys:
        warnings.warn(
            f"settings {sorted(unknown_keys)} are not supported."
        )

    def attribute_dict(methods: Dict[str, str]) -> Dict[str, ArgsKwargs]:
        """Create attribute dict according to attributes required in `settings`"""
        return {
            methods[name]: ctor_data
            for name, ctor_data in settings.items() if name in methods
        }

    struct_methods = attribute_dict(struct_method_mapping)
    extra_methods = attribute_dict(extra_method_mapping)

    base_message = "argument `base` must be a type derived from `System`"

    if base is None:
        base = System
    elif not inspect.isclass(base):
        raise TypeError(
            f"{base_message}; got {base!r}."
        )
    elif not issubclass(base, System):
        raise ValueError(
            f"{base_message}; got class `{base.__name__}`."
        )

    class Prototype(base):
        def setup(self, **kwargs):
            super().setup(**kwargs)
            def add_attributes(method_dict: dict):
                for method_name, values in method_dict.items():
                    if values is None:
                        continue
                    if not isinstance(values, list):
                        values = [values]
                    for info in values:
                        try:
                            args, kwargs = info  # expects a list of (tuple, dict)
                        except ValueError:
                            args, kwargs = [info], {}  # fallback
                        getattr(self, method_name)(*args, **kwargs)
            # Add inputs, outputs, transients, etc.
            add_attributes(struct_methods)
            # Add unknowns, equations & design methods
            add_attributes(extra_methods)
    
    return type(classname, (Prototype,), {})



from cosapp.systems import System
from cosapp.ports.port import BasePort
from cosapp.core.variableref import VariableReference
from cosapp.utils.naming import natural_varname
from cosapp.utils.helpers import check_arg

from typing import Dict, Set, Tuple


def get_free_inputs(system: System) -> Dict[str, VariableReference]:
    """Searches all free inputs within `system`, and returns a dictionary
    mapping each input to their source variable (may differ from input,
    in case of pulling).

    Returns:
    --------
    mapping: Dict[str, str]
        Dictionary of the kind `input_name`: `free_input_name`.
    """
    check_arg(system, 'system', System)

    def recursive_search(system: System, head: System) -> Tuple[Dict[str, str], Set[str]]:
        """Recursive inner version of `get_dependent_connections`"""
        inputs = dict()
        bound_inputs = set()  # type: Set[str]
        prefix = "" if system is head else f"{head.get_path_to_child(system)}."

        def get_portname(port: BasePort) -> str:
            if port.owner is system:
                return f"{prefix}{port.name}"  # avoid unnecessary path search
            return port.full_name(trim_root=True)

        for port in system.inputs.values():
            portname = get_portname(port)
            inputs.update({
                f"{portname}.{name}": f"{portname}.{name}"
                for name in port
            })

        for connector in system.incoming_connectors():
            if not connector.is_active:
                continue
            sink, source = connector.sink, connector.source
            if sink.is_output:
                continue
            sink_name = get_portname(sink)
            if source.is_output:
                bound_inputs |= set(
                    f"{sink_name}.{target}"
                    for target in connector.sink_variables()
                )
            else:
                source_name = get_portname(source)
                inputs.update({
                    f"{sink_name}.{target}": f"{source_name}.{origin}"
                    for target, origin in connector.mapping.items()
                })

        for child in system.children.values():
            child_inputs, child_bound_inputs = recursive_search(child, head)
            bound_inputs |= child_bound_inputs
            for key, alias in child_inputs.items():
                try:
                    # Check if `ref` is already mapped in parent
                    inputs[key] = inputs[alias]
                except KeyError:
                    inputs[key] = alias

        return inputs, bound_inputs

    root = system.root()
    
    if system is root or system.parent is root:
        # Most likely case
        start = system
    else:
        # If system is neither root nor a child of root,
        # the search must start just below root, to make sure
        # highest-level aliases are not connected to an output.
        # This is slow, and should be avoided.
        start = system.path()[1]

    inputs, bound_inputs = recursive_search(start, root)
    bound_inputs &= set(inputs.values())

    name2variable = root.name2variable
    result = dict()

    for key, src in inputs.items():
        if src in bound_inputs:
            continue
        ref = name2variable[key]
        try:
            context = system.get_path_to_child(ref.context)
        except:
            # context is not in system tree - skip key
            continue
        new_key = f"{ref.mapping.name}.{ref.key}"
        if context:
            new_key = f"{context}.{new_key}"
        result[natural_varname(new_key)] = result[new_key] = name2variable[src]

    return result



"""Systems from tutorials used in regression tests"""

from cosapp.systems import System
from cosapp.ports import Port
from typing import List


class Voltage(Port):
    def setup(self):
        self.add_variable('V', unit='V')

        
class Current(Port):
    def setup(self):
        self.add_variable('I', unit='A')


class Dipole(System):
    """Simple dipole model of the kind: {V_in, V_out} -> I"""
    def setup(self):
        self.add_input(Voltage, "V_in")
        self.add_input(Voltage, "V_out")
        self.add_output(Current, "I")
        self.add_outward("deltaV", 0.0)

    def compute(self):
        self.deltaV = self.V_in.V - self.V_out.V
        self.I.I = 0.0


class Resistor(Dipole):
    """Linear resistor model: {V_in, V_out} -> I"""
    def setup(self, R=100):
        super().setup()
        self.add_inward("R", float(R), unit="ohm", desc="Resistance")

    def compute(self):
        super().compute()
        self.I.I = self.deltaV / self.R


class Node(System):
    """Electric node model from Kirchhoff laws"""

    def setup(self, n_in=1, n_out=1):
        self.add_property('n_in', int(n_in))
        self.add_property('n_out', int(n_out))

        if min(self.n_in, self.n_out) < 1:
            raise ValueError("Node needs at least one incoming and one outgoing current")

        self.add_property('incoming', [
            self.add_input(Current, f"I_in{i}")
            for i in range(self.n_in)
        ])
        self.add_property('outgoing', [
            self.add_input(Current, f"I_out{i}")
            for i in range(self.n_out)
        ])
        
        self.add_inward('V', unit='V')
        self.add_outward('sum_I_in', 0., desc='Total incoming current')
        self.add_outward('sum_I_out', 0., desc='Total outgoing current')
        
        # Off-design problem
        self.add_unknown('V').add_equation(
            'sum_I_in == sum_I_out',
            name='Current balance',
        )
        
        self.add_inward('V_design', 10.)

    def compute(self):
        self.sum_I_in = sum(current.I for current in self.incoming)
        self.sum_I_out = sum(current.I for current in self.outgoing)

    @classmethod
    def make(cls, parent, name, incoming: List[Dipole], outgoing: List[Dipole], pulling=None) -> "Node":
        """Factory method making appropriate connections with parent system"""
        node = cls(name, n_in=max(len(incoming), 1), n_out=max(len(outgoing), 1))
        parent.add_child(node, pulling=pulling)
        
        for i, dipole in enumerate(incoming):
            parent.connect(dipole.V_out, node.inwards, "V")
            parent.connect(dipole.I, node[f"I_in{i}"])
        
        for i, dipole in enumerate(outgoing):
            parent.connect(dipole.V_in, node.inwards, "V")
            parent.connect(dipole.I, node[f"I_out{i}"])

        return node


class Source(System):
    def setup(self, I=0.1):
        self.add_inward('I', I, unit='A')
        self.add_output(Current, 'I_out', {'I': I})
    
    def compute(self):
        self.I_out.I = self.I


class Ground(System):
    def setup(self, V=0.0):
        self.add_inward('V', V, unit='V')
        self.add_output(Voltage, 'V_out', {'V': V})
    
    def compute(self):
        self.V_out.V = self.V



import abc
import logging, warnings
from numpy import bool_ as numpy_bool
from typing import Any, Union, Optional

from .zeroCrossing import ZeroCrossing
from cosapp.core.eval_str import EvalString
from cosapp.utils.naming import NameChecker, CommonPorts
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class EventState(abc.ABC):
    """Interface describing the inner state of an event
    """
    @abc.abstractmethod
    def must_emit(self) -> bool:
        pass

    def to_emit(self) -> bool:
        return False

    def value(self) -> Any:
        """Returns value associated with event.
        By default, equivalent to `must_emit()`.
        """
        return self.must_emit()

    @property
    def is_primitive(self) -> bool:
        """bool: `True` if event triggering is self contained, `False` otherwise"""
        return False

    def reset(self) -> None:
        pass

    def reevaluate(self) -> None:
        pass

    def tick(self) -> None:
        pass

    def lock(self) -> None:
        pass


class Event:
    # TODO: Doc!
    """Class for events, to be used as local variables"""

    __slots__ = (
        "_name",
        "_desc",
        "_context",
        "_present",
        "_trigger",
        "_state",
        "_final",
    )
    __name_check = NameChecker(excluded=CommonPorts.names())

    def __init__(
        self,
        name: str,
        context: "cosapp.systems.System",
        desc: str = "",
        trigger: Optional[Union[str, ZeroCrossing, EventState, "Event"]] = None,
        final: bool = False,
    ):
        """`Event` constructor.

        Parameters
        ----------
        - name [str]:
            Event name
        - context [System]:
            Multimode system in which event is defined.
        - desc [str, optional]:
            Description of the event (default: "")
        - trigger [Union[str, ZeroCrossing, EventState, Event], optional]:
            Trigger defining event occurrence, given as either a string, a `ZeroCrossing` object,
            another event, or an `EvenState` derived from another event.
            If set to `None` or absent (default), the event is undefined, and thus never occurs
            (at least until `trigger` is redefined).
        - final [bool, optional]:
            Defines whether the occurrence of the event should stop time simulations. Defaults to `False`.
        """
        self._name = self.name_check(name)

        from cosapp.systems import System
        check_arg(context, "context", System)
        self._context: System = context
        
        self._desc = desc
        self._present = False  # presence at current instant
        self._state: EventState = None
        self.trigger = trigger
        self.final = final

    @classmethod
    def name_check(cls, name: str):
        return cls.__name_check(name)

    @property
    def name(self) -> str:
        """str : Event name"""
        return self._name

    @property
    def contextual_name(self) -> str:
        """str : Join context system name and event name.

        If the event has no context, only its name is returned.
        """
        context = self._context
        return self._name if context is None else f"{context.name}.{self._name}"

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.contextual_name}>"

    def full_name(self, trim_root=False) -> str:
        """Returns full name up to root context.
        
        Parameters
        ----------
        trim_root : bool (optional, default False)
            Exclude root context name if True.

        Returns
        -------
        str
            The event full name
        """
        context = self._context
        path = []
        if context is not None:
            path = context.path_namelist()
            if trim_root:
                path = path[1:]
        path.append(self.name)
        return ".".join(path)

    @property
    def context(self) -> "cosapp.systems.System":
        return self._context

    @property
    def desc(self) -> str:
        """str : Event description"""
        return self._desc

    @property
    def trigger(self) -> Union[ZeroCrossing, EventState, "Event"]:
        return self._trigger

    @trigger.setter
    def trigger(self, trigger: Union[EventState, ZeroCrossing, str, "Event"]):
        state = None
        if isinstance(trigger, str):
            trigger = ZeroCrossing.from_comparison(trigger)
        if isinstance(trigger, ZeroCrossing):
            state = ZeroCrossingEvent(self, trigger)
        elif isinstance(trigger, Event):
            state = SynchronizedEvent(trigger)
        elif isinstance(trigger, EventState):
            state = trigger
        elif trigger is not None:
            raise TypeError(
                f"Event trigger cannot be defined by a {type(trigger).__name__}"
            )
        self._trigger = trigger
        self._state = state or UndefinedEvent()

    @property
    def is_primitive(self) -> bool:
        return self._state.is_primitive

    @property
    def present(self) -> bool:
        return self._present

    @property
    def final(self) -> bool:
        return self._final

    @final.setter
    def final(self, final: bool) -> None:
        check_arg(final, 'final', bool)
        self._final = final

    def value(self) -> Any:
        """Returns the value associated with the event."""
        return self._state.value()

    def __emit(self) -> None:
        self._present = True

    def tick(self) -> None:
        """Ticks the event, and locks it if it is a primitive event that has just been triggered."""
        self._state.tick()
        if self._present and self.is_primitive:
            self._state.lock()
        self._present = False

    def reset(self) -> None:
        """Resets the event."""
        self._state.reset()
        self._present = False

    def reevaluate(self) -> None:
        """Reevaluates the current state of the event; used to update information
        about zero-crossing events after an integration time step was interrupted by
        the triggering of an event."""
        self._state.reevaluate()

    def step(self) -> bool:
        """bool : Indicates whether the event was just triggered.
        
        Performs a step."""
        old = self._present
        if self._state.must_emit():
            self.__emit()
        # return True whenever the event has been emitted in this step
        return old ^ self._present

    def to_trigger(self) -> bool:
        """bool : Indicates whether the event has to be triggered in the next discrete step"""
        return self._present ^ self._state.to_emit()

    def filter(self, condition: str) -> "FilteredEvent":
        """Filters event with an additional boolean condition.
        
        Parameters:
        -----------
        - condition [str]:
            Evaluable boolean condition.
        
        Returns:
        --------
        - trigger [FilteredEvent]:
            The filtered event state, to be used as trigger.
        """
        return FilteredEvent(self, condition)

    @staticmethod
    def merge(*events: "Event") -> "MergedEvents":
        """Merges events into a trigger condition.
        
        Parameters:
        -----------
        - *events [Event]:
            Enumeration of events to be merged.
        
        Returns:
        --------
        - trigger [MergedEvents]:
            The merged event state, to be used as trigger.
        """
        return MergedEvents(*events)


class UndefinedEvent(EventState):
    """Inner state of an undefined, never occurring event"""
    def must_emit(self) -> bool:
        return False


class ZeroCrossingEvent(EventState):
    """Inner state of an event triggered by a zero-crossing expression"""
    def __init__(self, event: Event, zeroxing: ZeroCrossing):
        check_arg(event, 'event', Event)
        expr = EvalString(zeroxing.expression, event.context)
        if expr.constant:
            raise ValueError(f"Zero-crossing function {expr} is constant")
        if not isinstance(expr.eval(), float):
            raise TypeError(
                "Zero-crossing condition must be a float expression"
            )
        self._expr = expr
        self._direction = zeroxing.direction
        self._event = event
        self.reset()

    def reset(self) -> None:
        self._prev = self._curr = None
        self._locked = False

    def value(self) -> float:
        """Evaluates and returns the zero-crossing function defining the event."""
        return self._expr.eval()

    def zero_detected(self, next_value) -> bool:
        """Is a zero detected between previous value and `next_value`?"""
        return self._prev is not None and self._direction.zero_detected(self._prev, next_value)

    def must_emit(self) -> bool:
        """Checks whether the event is triggered in the current discrete step."""
        if self._curr is None and not self._event.present:
            self._curr = self.value()
            return not self._locked and self.zero_detected(self._curr)
        return False

    def to_emit(self) -> bool:
        """Checks whether the event will have to be triggered in a new discrete step"""
        if not self._locked and self._curr is None and not self._event.present:
            next_value = self.value()
            return self.zero_detected(next_value)
        return False
        
    def lock(self) -> None:
        """Locks the event"""
        self._locked = True

    def reevaluate(self):
        """Forces the reevaluation of the zero-crossing function"""
        self._curr = self.value()

    def tick(self):
        """Performs a tick and checks whether the event can be unlocked"""
        if self._locked:
            self._locked = (self._curr == self._prev)
        self._prev = self._curr
        self._curr = None

    def is_primitive(self) -> bool:
        return True


class FilteredEvent(EventState):
    """Inner state of an event triggered by another event,
    filtered by a Boolean condition.
    """
    def __init__(self, event: Event, condition: str):
        """`FilteredEvent` constructor.
        
        Parameters
        ----------
        event : Event
            Base event
        condition : str
            Boolean expression, as a string
        """
        check_arg(event, 'event', Event)
        expr = EvalString(condition, event.context)
        if not isinstance(expr.eval(), (bool, numpy_bool)):
            raise TypeError(
                "FilteredEvent condition must be a Boolean expression."
            )
        if expr.constant and not expr.eval():
            warnings.warn(
                f"Event {event.contextual_name} is filtered with"
                f" unconditionally false expression {str(expr)!r}",
                RuntimeWarning,
            )
        self._event = event
        self._condition = expr

    def must_emit(self) -> bool:
        return self._event.present and self._condition.eval()


class MergedEvents(EventState):
    """Inner state of an event triggered by the merging of
    of other, external events.
    """
    def __init__(self, *events: Event):
        """`MergedEvent` constructor.
        
        Parameters
        ----------
        *events : Event
            Any number of events
        """
        self._events = events

    def must_emit(self) -> bool:
        """Returns `True` if at least one event is present,
        `False` otherwise."""
        return any(event.present for event in self._events)


class SynchronizedEvent(EventState):
    """Inner state of an event synchronized with another event.
    """
    def __init__(self, event: Event):
        check_arg(event, 'event', Event)
        self._event = event

    def must_emit(self) -> bool:
        return self._event.present



# Copyright (C) 2023 twiinIT - All Rights Reserved
# Licensed under the Apache License, Version 2.0

import copy
import logging
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


def swap_system(old_system: "System", new_system: "System", init_values=True):
    """Replace `old_system` by `new_system`.

    Parameters
    ----------
    - old_system [System]: System to replace
    - new_system [System]: Replacement system
    - init_values [bool, optional]: If `True` (default), original
        system values are copied into the replacement system.
    
    Returns
    -------
    old_system [System]: the original system, devoid of parent.
    """
    from cosapp.base import System
    from cosapp.ports.port import BasePort

    check_arg(old_system, "old_system", System)
    check_arg(new_system, "new_system", System)

    if new_system.parent is not None:
        raise ValueError(
            f"System {new_system.full_name()!r} already belongs to a system tree."
        )
    if old_system.parent is None:
        raise ValueError(
            f"Cannot replace top system {old_system.full_name()!r}."
        )
    if new_system.name != old_system.name:
        logger.info(
            f"New system {new_system.name!r} renamed into {old_system.name!r} inside {old_system.parent.full_name()!r}"
        )

    new_system.parent = parent = old_system.parent
    new_system.name = system_name = old_system.name

    # connections list
    to_restore = list(
        filter(
            lambda c: (c.source.owner is old_system) or (c.sink.owner is old_system),
            parent.all_connectors(),
        )
    )

    # update child in parent
    execution_idx = list(parent.exec_order).index(system_name)
    parent.pop_child(system_name)
    parent.add_child(new_system, execution_index=execution_idx)

    # restore connections
    for connector in to_restore:
        sink = connector.sink
        source = connector.source
        if sink.owner is old_system:
            connector.sink = new_system[sink.name]
        if source.owner is old_system:
            connector.source = new_system[source.name]
        parent.connect(connector.sink, connector.source, connector.mapping)

    # init values
    if init_values:
        not_copied = set()
        for name, old_variable in old_system.name2variable.items():
            if not isinstance(old_variable.mapping, BasePort):
                continue
            try:
                new_variable = new_system.name2variable[name]
            except KeyError:
                not_copied.add(old_variable.contextual_name)
                continue
            try:
                new_variable.value = copy.deepcopy(old_variable.value)
            except:
                not_copied.add(old_variable.contextual_name)

        if not_copied:
            logger.warning(
                f"Could not copy {sorted(not_copied)} into {new_system.full_name()!r}"
            )

    return old_system



import inspect
from cosapp.patterns import Proxy
from cosapp.ports.port import BasePort
from cosapp.ports.enum import PortType
from cosapp.ports.connectors import BaseConnector
from typing import Dict, Any


class SystemConnector(Proxy):
    """Connector proxy used in `System`
    """
    def __init__(self, connector: BaseConnector):
        self.check(connector)
        super().__init__(connector)
        self.__noise: Dict[str, Any] = {}
        self.source.touch()
        self.activate()

    @staticmethod
    def check(wrappee: Any) -> None:
        """Checks whether `wrappee` can be wrapped in a `SystemConnector`
        proxy; raises an exception if not.

        Parameters:
        -----------
        - wrappee [Any]:
            If wrappee is a class, check that it is a concrete implementation of
            `BaseConnector`. If it is an object, check that its type is derived from
            `BaseConnector`.
        
        Raises:
        -------
        - `ValueError` if `wrappee` is a class not derived from `BaseConnector`
        - `TypeError` if `wrappee` is an object not derived from `BaseConnector`
        """
        if inspect.isclass(wrappee):
            ok = lambda t, base: issubclass(t, base) and t is not base
            error = ValueError
        else:
            ok = isinstance
            error = TypeError
        if not ok(wrappee, BaseConnector):
            raise error(
                "`SystemConnector` can only wrap objects derived from `BaseConnector`"
            )

    @property
    def is_active(self) -> bool:
        """bool: `True` if connector is activated, `False` otherwise."""
        return self.__is_active

    def activate(self) -> None:
        """Activate connector transfer."""
        self.__is_active = True

    def deactivate(self) -> None:
        """Deactivate connector transfer."""
        self.__is_active = False

    def set_perturbation(self, name: str, value: Any) -> None:
        """Add a perturbation on a connector.
        
        Parameters
        ----------
        name : str
            Name of the sink variable to perturb
        value : Any
            Perturbation value
        """
        if name not in self.sink_variables():
            raise ValueError(
                "Perturbations can only be added on sink variables"
            )
        self.__noise[name] = value
        self.sink.owner.touch()

    def clear_noise(self) -> None:
        self.__noise.clear()

    def __repr__(self):
        return repr(self.__wrapped__)
    
    def transfer(self) -> None:
        """Transfer values from `source` to `sink`."""
        if self.__is_active:
            sink: BasePort = self.sink
            source: BasePort = self.source
            noise = self.__noise

            if not source.is_clean or noise:
                sink.touch()
                self.__wrapped__.transfer()
                for varname, perturbation in noise.items():
                    sink[varname] += perturbation



import pytest

from cosapp.systems import System
from cosapp.drivers import NonLinearSolver, RunSingleCase


@pytest.fixture
def PressureLossFactory(test_library, test_data):
    def factory(suffix=""):
        return System.load(test_data / f"system_config_pressureloss{suffix}.json")
    return factory


@pytest.fixture
def system0(PressureLossFactory):
    """pressureLoss0D"""
    return PressureLossFactory()


@pytest.fixture
def system1(PressureLossFactory):
    """pressureLossSys (pressureLoss0D)"""
    s = PressureLossFactory(suffix="1")
    s.exec_order = ["p1"]
    return s


@pytest.fixture
def system2(PressureLossFactory):
    """pressureLossSys (pressureLoss0D + pressureLoss0D)"""
    s = PressureLossFactory(suffix="2")
    s.exec_order = ("p1", "p2")
    return s


@pytest.fixture
def system2tank(PressureLossFactory):
    """pressureLossSys (pressureLoss0D + tank + pressureLoss0D a pressureLossSys)"""
    s = PressureLossFactory(suffix="2tank")
    s.exec_order = ("p1", "tank1", "p2")
    return s


@pytest.fixture
def system11(PressureLossFactory):
    """pressureLossSys(pressureLoss0D) with direct connexion of module
    """
    s = PressureLossFactory(suffix="11")
    s.exec_order = ["p1"]
    s.p1.exec_order = ["p11"]
    return s


@pytest.fixture
def system11bis(PressureLossFactory):
    """pressureLossSys(pressureLoss0D) with connexion of module using `..`
    """
    return PressureLossFactory(suffix="11bis")


@pytest.fixture
def system12(PressureLossFactory):
    """pressureLossSys(pressureLoss0D + pressureLoss0D)
    """
    s = PressureLossFactory(suffix="12")
    s.exec_order = ["p1"]
    s.p1.exec_order = ["p11", "p12"]
    return s


@pytest.fixture
def system22(PressureLossFactory):
    """pressureLossSys(pressureLoss0D + pressureLoss0D) + pressureLoss0D
    """
    s = PressureLossFactory(suffix="22")
    s.exec_order = ["p1", "p2"]
    s.p1.exec_order = ["p11", "p12"]
    return s


@pytest.fixture
def system121(PressureLossFactory):
    """pressureLoss0D + Splitter12 + pressureLoss0D//pressureLoss0D + Mixer21 pressureLoss0D
    """
    s = PressureLossFactory(suffix="121")
    s.exec_order = "p1", "sp", "p2", "p3", "mx", "p4"
    return s


@pytest.fixture
def system131(PressureLossFactory):
    """pressureLoss0D + Mixer21 + pressureLoss0D//pressureLoss0D(backward) + Splitter12 + pressureLoss0D
    """
    s = PressureLossFactory(suffix="131")
    s.exec_order = "p1", "mx", "p2", "sp", "p3", "p4"
    s.p3.K = -12100
    s.p2.K = 100
    s.sp.x = 0.5
    return s


@pytest.fixture
def system222(PressureLossFactory):
    """pLS(pl0 + mx + (pLS (pL0 + mx + (pL0+pL0//pl0(b) + sp + pl0))//pl0(b)) + sp + pL0)
    """
    s = PressureLossFactory(suffix="222")
    s.exec_order = "p1", "m1", "p2", "s1", "p3", "p4"
    s.flnum_in.W = 9.0
    s.flnum_in.Pt = 108100
    s.p2.p22.p221.K = 50
    s.p2.p22.p222.K = 50
    s.p3.K = -32100
    s.p2.p23.K = -12100
    return s


def test_system0(system0):
    s = system0
    s.run_once()
    assert s.flnum_out.Pt == 90000.0


def test_system1(system1):
    s = system1
    s.run_once()
    assert s.p1.flnum_out.Pt == 90000.0


def test_system2(system2):
    s = system2
    s.run_once()
    assert s.p2.flnum_out.Pt == 80000.0


def test_system2tank(system2tank):
    s = system2tank

    # tank is half full
    s.tank1.vol = 10.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 89900.0
    assert s.tank1.vol == 10.9
    assert s.tank1.flnum_out.W == 1.0

    # tank is empty
    s.tank1.vol = 0.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 89900.0
    assert s.tank1.vol == 0.9
    assert s.tank1.flnum_out.W == 1.0

    # tank is full
    s.tank1.vol = 100.0
    s.run_once()
    assert s.p2.flnum_out.Pt == 80000.0
    assert s.tank1.vol == pytest.approx(100)
    assert s.tank1.flnum_out.W == 10.0


def test_system11(system11):
    s = system11
    s.run_once()
    assert s.p1.p11.flnum_out.Pt == pytest.approx(90000, rel=5e-8)


def test_system11bis(system11bis):
    s = system11bis
    s.run_once()
    assert s.p2.p21.flnum_out.Pt == pytest.approx(80000, rel=5e-8)
    assert s.p3.flnum_out.Pt == pytest.approx(70000, rel=5e-8)


def test_system12(system12):
    s = system12
    s.run_once()
    assert s.p1.p12.flnum_out.Pt == pytest.approx(80000, rel=5.0e-8)


def test_system22(system22):
    s = system22
    s.run_once()
    assert s.p2.flnum_out.Pt == pytest.approx(70000, rel=5.0e-8)


def test_system121(system121: System):
    s = system121
    # test sub-systems are listed in the execution order list
    assert list(s.children) == list(s.exec_order)

    s.run_once()
    assert s.p4.flnum_out.Pt == 74800.0

    s.add_driver(NonLinearSolver("design", tol=5.0e-8))
    s.run_drivers()

    # test iterative loop detection
    assert len(s.residues) == 0
    problem = s.assembled_problem()
    assert set(problem.residues) == {"mx: epsilon == 0"}

    assert s.p4.flnum_out.Pt == pytest.approx(77500, rel=1e-6)
    assert s.p2.flnum_out.W == pytest.approx(5, rel=1e-6)
    assert s.mx.flnum_out.W == pytest.approx(s.mx.flnum_in1.W + s.mx.flnum_in2.W, rel=1e-5)


def test_system121_redundant_1(system121):
    """Same as `test_system121` with redundant declaration of unknowns.

    Case 1: system unknown `sp.x` is redeclared as unknown at solver level.
    """
    s = system121
    d = s.add_driver(NonLinearSolver("solver"))
    d.add_unknown("sp.x")

    with pytest.raises(ValueError, match="'sp\.x' is defined as design and off-design unknown"):
        s.run_drivers()


def test_system121_redundant_2(system121):
    """Same as `test_system121` with redundant declaration of unknowns.

    Case 2: system unknown `sp.x` is redeclared as unknown at runner level.
    """
    s = system121
    solver = s.add_driver(NonLinearSolver("solver"))
    solver.add_child(RunSingleCase("case"))
    solver.case.add_unknown("sp.x")

    with pytest.raises(ValueError, match="'sp\.x' already exists in 'offdesign'"):
        s.run_drivers()


def test_system131(system131):
    # ! Do not suppress this unit test, it's the only one doing new computation with resetting the system
    s = system131
    d = s.add_driver(NonLinearSolver("solver"))

    s.run_drivers()
    assert s.p4.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(11, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 11, rel=2e-5)

    s.p3.K = -3600
    s.run_drivers()
    assert s.p4.flnum_out.Pt == pytest.approx(65600, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(12, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 12, rel=1e-5)

    s.p3.K = -12100
    d.options["factor"] = 0.7
    s.run_drivers()
    assert s.p4.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(11, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 11, rel=1e-5)

    s.p3.K = -260100
    s.run_drivers()

    assert s.p2.flnum_out.W == pytest.approx(10.2, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(10.0, rel=1e-5)
    assert s.sp.x == pytest.approx(10 / 10.2, rel=1e-5)
    assert s.p4.flnum_out.Pt == pytest.approx(69596, rel=1e-5)


def test_system222(system222):
    """pLS(pl0 + mx + (pLS (pL0 + mx + (pL0+pL0//pl0(b) + sp + pl0))//pl0(b)) + sp + pL0)
    """
    s = system222
    d = s.add_driver(NonLinearSolver("solver"))
    s.run_drivers()

    assert s.p2.p24.flnum_out.Pt == pytest.approx(67900, rel=1e-5)
    assert s.p2.flnum_out.W == pytest.approx(10, rel=1e-5)
    assert s.p4.flnum_out.W == pytest.approx(9, rel=1e-5)
    assert s.s1.x == pytest.approx(9 / 10, rel=1e-5)
    assert s.p2.s21.x == pytest.approx(10 / 11, rel=1e-5)


def test_system222_redundant(system222):
    """Same as `test_system222` with redundant off-design unknowns.
    """
    s = system222
    solver = s.add_driver(NonLinearSolver("solver"))
    solver.add_child(RunSingleCase("case"))
    solver.case.add_unknown(["p2.s21.x", "s1.x"])

    with pytest.raises(ValueError, match="'p2\.s21\.x' already exists in 'offdesign'"):
        s.run_drivers()



"""Dummy call on D3 views"""
import pytest

from cosapp.tests.library.systems import FanComplex
from cosapp.tools.views.d3js import  D3JsRenderer
from cosapp.utils.testing import no_exception, assert_keys

pytest.importorskip("jinja2")


@pytest.fixture(scope="function")
def fan():
    return FanComplex("fan")


@pytest.fixture(scope="function")
def renderer(fan):
    return D3JsRenderer(fan)


def test_D3JsRenderer_get_level(fan):
    level = D3JsRenderer.get_level(fan)
    assert level == 3


def test_D3JsRenderer_html_tags(renderer):
    tags = renderer.html_tags()
    assert_keys(
        tags,
        "html_begin_tags",
        "html_end_tags",
    )


def test_D3JsRenderer_html_resources(renderer):
    resources = renderer.html_resources()
    assert_keys(resources, "d3_js", "draw_js", "d3_styles")


def test_D3JsRenderer_html_template(renderer):
    with no_exception():
        template = renderer.html_template()


def test_D3JsRenderer_get_globals(renderer):

    common = renderer.get_globals()
    assert_keys(
        common,
        "template",
        "html_begin_tags",
        "html_end_tags",
        "d3_js",
        "draw_js",
        "d3_styles",
    )


def test_D3JsRenderer_html_content(renderer):
    with no_exception():
        renderer.html_content()



import abc
import logging
import numpy
import pandas
import copy
from io import StringIO
from numbers import Number
from typing import Tuple, NamedTuple, List, Dict, Union, Optional, Any

from cosapp.core.time import UniversalClock
from cosapp.ports.enum import PortType
from cosapp.drivers.driver import Driver, Recorder
from cosapp.drivers.time.utils import (
    TimeUnknown,
    TimeVarManager,
    TimeStepManager,
    TwoPointCubicInterpolator,
)
from cosapp.drivers.time.scenario import Scenario
from cosapp.multimode.discreteStepper import DiscreteStepper
from cosapp.multimode.event import Event
from cosapp.recorders import DataFrameRecorder
from cosapp.utils.helpers import check_arg
from cosapp.utils.logging import LogFormat, LogLevel

logger = logging.getLogger(__name__)


class EventRecord(NamedTuple):
    """Named tuple associating a list of joint events
    and their occurrence time.
    """
    time: float
    events: List[Event]


class ExplicitTimeDriver(Driver):
    """
    Generic implementation of an explicit time driver with constant time step.
    Specialization of derived classes is achieved by the implementation of abstract method `_update_transients`
    """

    __slots__ = (
        '__time_interval', '__clock', '__recordPeriod',
        '_transients', '_rates', '__dt_manager', '__var_manager',
        '__scenario', 'record_dt', '__recorded_dt', '__stepper',
        '__event_data', '__recorded_events',
    )

    def __init__(self,
        name = "Explicit time driver",
        owner: Optional["cosapp.systems.System"] = None,
        time_interval: Tuple[float, float] = None,
        dt: float = None,
        record_dt: bool = False,
        **options):
        """Initialization of the driver

        Parameters
        ----------
        name : str, optional
            Driver's name; default: "Explicit time driver"
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belongs; default None
        time_interval : Tuple[float, float]
            Time interval [t_begin, t_end], with t_end > t_begin; defaut None
        dt : float
            Time step; defaut None. If None, will be tentatively determined
            from system transient variables.
        record_dt : bool
            If True, driver will store actual time steps used in simulation; default False.
            This option is only useful for post-run analysis, when `dt` is unspecified.
        **options : Dict[str, Any]
            Optional keywords arguments for generic `Driver` objects
        """
        dt_growth_rate = options.pop('max_dt_growth_rate', 2)
        super().__init__(name, owner, **options)
        self.__time_interval = None
        self.__recorded_dt = numpy.array([])
        self.__clock = UniversalClock()
        self.__recordPeriod = None
        self.__dt_manager = TimeStepManager(max_growth_rate=dt_growth_rate)
        self.__var_manager: TimeVarManager = None
        self._transients = dict()
        self._rates = dict()
        self.dt = dt
        self.time_interval = time_interval
        self.record_dt = record_dt
        self.__scenario = Scenario("empty", self)
        self.__stepper: DiscreteStepper = None
        self.__event_data: pandas.DataFrame = None
        self.__recorded_events: List[EventRecord] = []

    def is_standalone(self) -> bool:
        return True

    @property
    def dt(self) -> Union[None, Number]:
        """Nominal time step (None if unspecified)"""
        return self.__dt_manager.nominal_dt

    @dt.setter
    def dt(self, value: Number) -> None:
        self.__dt_manager.nominal_dt = value

    @property
    def time(self) -> Number:
        """Current simulation time"""
        return self.__clock.time

    @property
    def time_interval(self) -> Tuple[Number, Number]:
        """Time interval covered by driver"""
        return self.__time_interval

    @time_interval.setter
    def time_interval(self, interval: Tuple[Number, Number]) -> None:
        if interval is not None:
            check_arg(interval, 'time_interval', (tuple, list), lambda it: len(it) == 2)
            interval = tuple(interval)
            start, end = interval
            check_arg(start, 'start time', Number, value_ok = lambda t: t >= 0)
            check_arg(end, 'end time', Number, value_ok = lambda t: t >= start and numpy.isfinite(t))
        self.__time_interval = interval

    @property
    def event_data(self) -> pandas.DataFrame:
        """pandas.DataFrame: DataFrame detailing all event cascades occurring during simulation"""
        return self.__event_data

    @property
    def recorded_events(self) -> List[EventRecord]:
        """List[EventRecord]: list of recorded event cascades"""
        return self.__recorded_events

    # @property
    # def start_time(self):
    #     return self.__time_interval[0]

    # @property
    # def end_time(self):
    #     return self.__time_interval[1]

    def set_scenario(self,
        name = "scenario",
        init: Dict[str, Any] = dict(),
        values: Dict[str, Any] = dict(),
        stop: Optional[Union[str, Event]] = None,
    ) -> None:
        """
        Define a simulation scenario, from initial and boundary conditions.

        Parameters
        ----------
        init : dict
            Dictionary of initial conditions, of the kind {variable: value}
        values : dict
            Dictionary of boundary conditions, of the kind {variable: value}
            Explicit time dependency may be specified, as {variable: 'cos(omega * t)'}, for example
        name : str
            Name of the scenario (by default, 'scenario')
        """
        self.__scenario = scenario = Scenario.make(name, self, init, values)
        if stop is not None:
            scenario.stop.trigger = stop

    @property
    def scenario(self) -> Scenario:
        """Scenario: the simulation scenario, defining initial and boundary conditions"""
        return self.__scenario

    def add_recorder(self, recorder: Recorder, period: Optional[Number] = None) -> Recorder:
        """Add an internal recorder storing the time evolution of values of interest.

        Parameters
        ----------
        - recorder [BaseRecorder]:
            The recorder to be added.
        - period [Number, optional]:
            Recording period. If `None` (default), data are recorded at all time steps.
        """
        if 'time' not in recorder:
            cls = type(recorder)
            recorder = cls.extend(recorder, includes='time')
        super().add_recorder(recorder)
        if period is not None:
            self.recording_period = period
        return self.recorder

    def setup_run(self) -> None:
        """Setup the driver once before starting the simulation and before
        calling the systems `setup_run`.
        """
        if self.__time_interval is None:
            raise ValueError("Time interval was not specified")

        self.__stepper = DiscreteStepper(self)
        self.__var_manager = manager = TimeVarManager(self.owner)
        self.__dt_manager.transients = manager.transients
        self._transients = manager.transients
        self._rates = manager.rates
        logger.debug(f"Transient variables: {self._transients!r}")
        logger.debug(f"Rate variables: {self._rates!r}")
        self.__reset_time()

    def run_once(self) -> None:
        """Run time driver once, assuming driver has already been initialized.
        """
        with self.log_context(" - run_once"):
            if self.is_active():
                self._precompute()

                # Sub-drivers are executed at each time step in `compute`,
                # so the child loop before `self.compute()` is omitted.
                logger.debug(f"Call {self.name}.compute")
                self._compute_calls += 1
                self.compute()

                self._postcompute()
                self.computed.emit()

            else:
                logger.debug(f"Skip {self.name} execution - Inactive")

    def compute(self) -> None:
        """Simulate the time-evolution of owner System over a prescribed time interval"""
        self._initialize()

        t0, t_end = self.__time_interval
        dt_manager = self.__dt_manager
        dt = self.dt or dt_manager.time_step()
        record_all = False
        t_record = numpy.inf
        recorder = self._recorder

        if recorder is None:
            must_record = lambda t, t_record: False
            record_data = lambda *args, **kwargs: None
            event_rec_options = {}
        else:
            if self.recording_period is None:
                if self.dt is None:
                    logger.warning(
                        "Unspecified recording period and time step"
                        "; all time steps will be recorded"
                    )
                record_all = True
            eps = min(1e-8, dt / 100)
            must_record = lambda t, t_record: abs(t - t_record) < eps
            def record_data(stamp=None):
                if stamp is None:
                    stamp = f"t={float(self.time):.14}"
                recorder.record_state(stamp, self.status, self.error_code)
            t_record = numpy.inf if record_all else t0
            event_rec_options = dict(
                (attr, getattr(recorder,attr))
                for attr in (
                    "includes",
                    "excludes",
                    "section",
                    "precision",
                    "hold",
                )
            )
            event_rec_options["numerical_only"] = recorder._numerical_only
            event_rec_options["raw_output"] = recorder._raw_output #TODO: Expose both to the API?
        event_rec = DataFrameRecorder(**event_rec_options)
        event_rec.watched_object = self.owner
        def record_event(stamp=None):
            if stamp is None:
                stamp = f"t={float(self.time):.14}"
            event_rec.record_state(stamp, self.status, self.error_code)

        recorded_dt = []
        if self.record_dt:
            record_dt = lambda dt: recorded_dt.append(dt)
        else:
            record_dt = lambda dt: None
        
        t = t0
        n_record = 0
        prev_dt = None
        self._set_time(t0)

        stepper = self.__stepper
        manager = self.__var_manager
        owner_transients = manager.problem.transients
        self.__recorded_events = []

        def value_and_derivative(var: TimeUnknown):
            return (
                copy.copy(var.value),
                copy.copy(var.d_dt),
            )

        tr_data = dict(
            (key, [*value_and_derivative(var), None, None])
            for key, var in owner_transients.items()
        )

        def update_system(t, dt) -> float:
            """Continuously update owner system over one time step,
            and check for any event occurrence afterwards.
            """
            nonlocal tr_data
            self._update_transients(dt)
            self._set_time(t + dt)
            # Store transient values and derivatives @ t + dt
            for key, var in owner_transients.items():
                tr_data[key][2:4] = value_and_derivative(var)
            
            if stepper.event_detected():
                stepper.set_data(
                    interval = (t, t + dt),
                    interpol = {
                        name: TwoPointCubicInterpolator(
                            xs = (t, t + dt),
                            ys = data[0::2],  # values
                            dy = data[1::2],  # derivatives
                        )
                        for name, data in tr_data.items()
                    }
                )
                occur = stepper.first_discrete_step()  # first step: root finding + non-primitive events
                record_data()
                record_event()
                stepper.reevaluate_primitive_events()
                self.transition()
                record_event(occur.event.contextual_name)
                record = EventRecord(occur.time, [occur.event])
                all_events = set(stepper.present_events())
                stepper.tick()
                stepper.set_events()

                while stepper.event_detected(): # following steps: event cascade
                    events = stepper.discrete_step()
                    all_events.update(events)
                    self.transition()
                    stamp = ", ".join(event.contextual_name for event in events)
                    record_event(stamp)
                    stepper.tick()
                    stepper.set_events()

                record.events.extend(all_events - {occur.event})
                self.__recorded_events.append(record)
                for transient in self._transients.values():
                    transient.touch()
                for event in record.events:
                    event.context.set_dirty(PortType.IN)
                self._set_time(occur.time)
                self._synch_transients()
                record_data(occur.event.contextual_name)
                must_stop = any(event.final for event in all_events)
                next_t = occur.time
                dt = next_t - t

            else:
                next_t = t + dt
                stepper.reevaluate_primitive_events()
                stepper.shift()
                must_stop = False

            record_dt(dt)
            for data in tr_data.values():
                data[0:2] = data[2:4]

            return next_t, must_stop

        stopped = False

        while not stopped:  # time loop
            if record_all:
                record_data()
            elif must_record(t, t_record):
                record_data()
                n_record += 1
                t_record = t0 + n_record * self.recording_period
            dt = dt_manager.time_step(prev_dt)
            next_t = t + dt
            # Update previous dt, unless current dt is artificially
            # limited by recording timestamp `t_record`
            t_record = min(t_end, t_record)
            if next_t > t_record:
                next_t = t_record
                remaining = t_end - t
                if remaining < 1e-3 * dt:
                    break
                dt = next_t - t
            else:
                prev_dt = dt
            t, stopped = update_system(t, dt)

        self.__recorded_dt = numpy.asarray(recorded_dt)
        self.__event_data = event_rec.export_data()
        try:
            last_record = self.__recorded_events[-1]
        except IndexError:
            pass
        else:
            if self.__scenario.stop in last_record.events:
                logger.info(
                    f"Stop criterion met at t = {last_record.time}"
                )

    def transition(self) -> None:
        """Execute owner system transition and reinitialize sub-drivers"""
        self.owner.tree_transition()
        # Reinitialize sub-drivers after system transition
        for driver in self.children.values():
            driver.call_setup_run()

    def _set_time(self, t: Number) -> None:
        dt = t - self.time
        self.__clock.time = t
        self.__scenario.update_values()
        self._update_children()
        self._update_rates(dt)

    def __reset_time(self) -> None:
        """Reset clock time to driver start time"""
        time = self.__time_interval[0]
        logger.debug(f"Reset time to {time}")
        self.__clock.reset(time)

    def _initialize(self):
        self.__reset_time()
        self.__scenario.apply_init_values()
        self.__scenario.update_values()
        self.__init_modevars()
        self._synch_transients()
        logger.debug("Reset rates")
        for rate in self._rates.values():
            rate.reset()
        self.__stepper.reset()
    
    def __init_modevars(self):
        """Force init value of output mode vars or owner system."""
        from cosapp.systems import System
        
        for system in self.owner.tree():
            name = system.full_name()
            port = system[System.MODEVARS_OUT]
            for variable in port.variables():
                variable.initialize()
                logger.debug(
                    f"Mode variable {name}.{variable.name} set to {variable.value}"
                )

    def _synch_transients(self):
        """Re-synch stacked unknowns with root variables"""
        for transient in self._transients.values():
            transient.reset()

    def _update_children(self) -> None:
        """Execute sub-drivers, if any, or owner's subsystem drivers"""
        if self.children:
            for driver in self.children.values():
                driver.run_once()
        else:
            self.owner.run_children_drivers()

    def _update_rates(self, dt: Number) -> None:
        """Update rate-of-changes over time interval dt"""
        if dt == 0:
            return
        synch_needed = False
        for rate in self._rates.values():
            rate.update(dt)
            synch_needed = True
        if synch_needed:
            # Re-run dynamic system with updated (synchronized) rates
            # Equivalent to a single step fixed-point solver
            self._update_children()

    @abc.abstractmethod
    def _update_transients(self, dt: Number) -> None:
        """
        Time integration of transient variables over time step `dt`.
        Actual implementation depends on chosen numerical scheme.
        """
        pass

    @property
    def recording_period(self) -> Number:
        """float: Recording period of time driver's internal recorder"""
        return self.__recordPeriod

    @recording_period.setter
    def recording_period(self, value: Number):
        check_arg(value, 'recording_period', Number, lambda t: t > 0)
        interval = self.time_interval
        if interval is None:
            self.__recordPeriod = value
        else:
            self.__recordPeriod = min(value, interval[1] - interval[0])

    @property
    def recorded_dt(self) -> numpy.ndarray:
        """numpy.ndarray: list of time steps recorded during driver execution"""
        return self.__recorded_dt

    def log_debug_message(self, handler: "HandlerWithContextFilters", record: logging.LogRecord, format: LogFormat = LogFormat.RAW) -> bool:
        """Callback method on the system to log more detailed information.
        
        This method will be called by the log handler when :py:meth:`~cosapp.utils.logging.LoggerContext.log_context`
        is active if the logging level is lower or equals to VERBOSE_LEVEL. It allows
        the object to send additional log message to help debugging a simulation.

        Parameters
        ----------
        handler : HandlerWithContextFilters
            Log handler on which additional message should be published.
        record : logging.LogRecord
            Log record
        format : LogFormat
            Format of the message

        Returns
        -------
        bool
            Should the provided record be logged?
        """
        message = record.getMessage()
        activate = getattr(record, "activate", None)
        emit_record = super().log_debug_message(handler, record, format)

        if message.endswith("call_setup_run") or message.endswith("call_clean_run"):
            emit_record = False

        elif activate == True:
            self.record_dt = True
            emit_record = False

        elif activate == False:
            self.record_dt = False
            emit_record = False
            container = StringIO()
            numpy.savetxt(container, self.recorded_dt, delimiter=",")
            dts = container.getvalue()
            
            message = f"Time steps:\n{dts}"
            handler.log(
                LogLevel.FULL_DEBUG,
                message,
                name=logger.name,
            )

        return emit_record



import pytest

from cosapp.drivers import NonLinearMethods, NonLinearSolver
from cosapp.tests.library.systems import (
    SimpleTurbofan,
    AdvancedTurbofan,
    ComplexTurbofan,
)


@pytest.fixture
def simple_turbofan():
    engine = SimpleTurbofan("engine")

    engine.exec_order = ["atm", "inlet", "fan", "duct", "noz"]
    engine.duct.cst_loss = 0.98
    engine.fan.mech_in.XN = 80

    solver = engine.add_driver(NonLinearSolver("solver"))
    solver.add_unknown("fan.mech_in.PW")

    return engine


@pytest.fixture
def advanced_turbofan():
    engine = AdvancedTurbofan("engine")

    engine.exec_order = ["atm", "inlet", "fan", "merger", "duct", "bleed", "noz"]
    engine.duct.cst_loss = 0.98
    engine.fan.mech_in.XN = 80

    solver = engine.add_driver(NonLinearSolver("solver"))
    solver.add_unknown("fan.mech_in.PW")

    return engine


@pytest.fixture
def complex_turbofan():
    def factory(with_solver=False):
        engine = ComplexTurbofan("engine")
        engine.duct.cst_loss = 0.98
        engine.fanC.ductC.duct.cst_loss = 0.98
        engine.fanC.ductC.bleed.split_ratio = 0.995
        engine.fanC.mech_in.XN = 80
        engine.bleed.split_ratio = 0.99

        if with_solver:
            solver = engine.add_driver(NonLinearSolver("solver", tol=1e-7))
            solver.add_unknown("fanC.mech_in.PW")

        return engine

    return factory


def test_SimpleTurbofan_set_inputs(simple_turbofan):
    engine = simple_turbofan

    assert engine.fan.mech_in.XN == 80.0
    assert engine.duct.cst_loss == 0.98


def test_SimpleTurbofan_solve_80(simple_turbofan):
    """Solve simple turbofan @ XN = 80"""
    engine = simple_turbofan
    engine.fan.mech_in.XN = 80
    engine.run_drivers()

    assert engine.fan.mech_in.XN == 80.0
    assert engine.fan.gh == pytest.approx(0.0385792701, rel=1e-5)
    assert engine.fan.mech_in.PW == pytest.approx(8279856, rel=1e-5)
    # a second run changes fan power because of poor fsolve behavior
    # first run gives fan power residue bigger than 1e-5
    assert engine.inlet.W_in.W == pytest.approx(161.463, rel=1e-4)


def test_SimpleTurbofan_solve_100(simple_turbofan):
    """Solve simple turbofan @ XN = 100"""
    engine = simple_turbofan
    engine.fan.mech_in.XN = 100
    engine.run_drivers()

    assert engine.fan.mech_in.XN == 100.0
    assert engine.fan.gh == pytest.approx(0.1566921207, rel=1e-5)
    assert engine.fan.mech_in.PW == pytest.approx(10874232, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(177.034, rel=1e-5)


def test_AdvancedTurbofan_set_inputs(advanced_turbofan: AdvancedTurbofan):
    engine = advanced_turbofan

    assert engine.fan.mech_in.XN == 80.0
    assert engine.duct.cst_loss == 0.98


def test_AdvancedTurbofan_solve_80(advanced_turbofan: AdvancedTurbofan):
    """Solve advanced turbofan @ XN = 80"""
    engine = advanced_turbofan
    engine.fan.mech_in.XN = 80
    engine.run_drivers()

    assert engine.fan.mech_in.XN == 80.0
    assert engine.fan.gh == pytest.approx(0.038579270081290681, rel=1e-5)
    assert engine.fan.mech_in.PW == pytest.approx(8279856, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(161.463, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.630942952, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.630942952, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(180023, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(180023, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)


def test_AdvancedTurbofan_solve_100(advanced_turbofan: AdvancedTurbofan):
    """Solve advanced turbofan @ XN = 100"""
    engine = advanced_turbofan
    engine.fan.mech_in.XN = 100
    engine.run_drivers()

    assert engine.fan.mech_in.XN == 100.0
    assert engine.fan.gh == pytest.approx(0.15669212068943689, rel=1e-5)
    assert engine.fan.mech_in.PW == pytest.approx(10874232, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(177.034, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.7882221046, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.7882221046, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(200435.4, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(200435.4, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(334.330, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(334.330, rel=1e-5)


def test_ComplexTurbofan_set_inputs(complex_turbofan):
    engine: ComplexTurbofan = complex_turbofan(with_solver=False)

    assert engine.fanC.mech_in.XN == 80.0
    assert engine.duct.cst_loss == 0.98
    assert engine.fanC.ductC.duct.cst_loss == 0.98
    assert engine.fanC.ductC.bleed.split_ratio == 0.995
    assert engine.bleed.split_ratio == 0.99


def test_ComplexTurbofan_solve_80(complex_turbofan):
    """Solve complex turbofan @ XN = 80 (ref case).
    """
    engine: ComplexTurbofan = complex_turbofan(with_solver=True)
    engine.run_drivers()

    assert engine.fanC.mech_in.XN == 80.0
    assert engine.fanC.gh == pytest.approx(0.03857927008129068, rel=1e-5)
    assert engine.fanC.mech_in.PW == pytest.approx(8114258.93, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(158.2341, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)

    assert engine.fanC.ductC.bleed.fl2_out.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Tt == pytest.approx(273.15, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Tt == pytest.approx(273.15, rel=1e-5)


def test_ComplexTurbofan_inner_solver_ductC_solve_80(complex_turbofan):
    """Complex turbofan with local loop solved with an inner solver
    (subsystem `ductC`).
    """
    engine: ComplexTurbofan = complex_turbofan(with_solver=True)

    # Add inner driver to solve local loop
    engine.fanC.ductC.add_driver(NonLinearSolver("design", method=NonLinearMethods.POWELL))
    engine.run_drivers()

    assert engine.fanC.mech_in.XN == 80.0
    assert engine.fanC.gh == pytest.approx(0.03857927008, rel=1e-5)
    assert engine.fanC.mech_in.PW == pytest.approx(8114258.932, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(158.2341, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)

    assert engine.fanC.ductC.bleed.fl2_out.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Tt == pytest.approx(273.15, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Tt == pytest.approx(273.15, rel=1e-5)


def test_ComplexTurbofan_inner_solver_fanC_solve_80(complex_turbofan):
    """Complex turbofan with local loop solved with an inner solver
    (subsystem `fanC`).
    """
    engine: ComplexTurbofan = complex_turbofan(with_solver=False)
    engine.add_driver(NonLinearSolver("solver", tol=1e-7))

    # Add inner solver to solve local loop
    fan_solver = engine.fanC.add_driver(NonLinearSolver("solver", tol=1e-7))
    fan_solver.add_unknown("mech_in.PW")

    engine.run_drivers()

    assert engine.fanC.mech_in.XN == 80.0
    assert engine.fanC.gh == pytest.approx(0.03857927008, rel=1e-5)
    assert engine.fanC.mech_in.PW == pytest.approx(8114258.932, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(158.2341, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.59832409, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(176422.55, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(324.2258, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(324.2258, rel=1e-5)

    assert engine.fanC.ductC.bleed.fl2_out.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.W == pytest.approx(0.79514614358, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Pt == pytest.approx(98802.0075, rel=1e-5)
    assert engine.fanC.ductC.bleed.fl2_out.Tt == pytest.approx(273.15, rel=1e-5)
    assert engine.fanC.ductC.merger.fl2_in.Tt == pytest.approx(273.15, rel=1e-5)


def test_ComplexTurbofan_solve_100(complex_turbofan):
    """Solve complex turbofan @ XN = 100.
    """
    engine: ComplexTurbofan = complex_turbofan(with_solver=True)
    engine.fanC.mech_in.XN = 100.0
    engine.run_drivers()

    assert engine.fanC.mech_in.XN == 100.0
    assert engine.fanC.gh == pytest.approx(0.15669212069, rel=1e-5)
    assert engine.fanC.mech_in.PW == pytest.approx(10656747.6, rel=1e-5)
    assert engine.inlet.W_in.W == pytest.approx(173.4933, rel=1e-5)

    assert engine.bleed.fl2_out.W == pytest.approx(1.75245766, rel=1e-5)
    assert engine.merger.fl2_in.W == pytest.approx(1.75245766, rel=1e-5)
    assert engine.bleed.fl2_out.Pt == pytest.approx(196426.7, rel=1e-5)
    assert engine.merger.fl2_in.Pt == pytest.approx(196426.7, rel=1e-5)
    assert engine.bleed.fl2_out.Tt == pytest.approx(334.330, rel=1e-5)
    assert engine.merger.fl2_in.Tt == pytest.approx(334.330, rel=1e-5)


def test_ComplexTurbofan_to_html(tmp_path, complex_turbofan):
    engine: ComplexTurbofan = complex_turbofan(with_solver=True)
    engine.to_html(tmp_path / "complexturbofan.html")



"""Classes and functions helper using third-party libraries to interact with CoSApp core.
"""
from .logging import LogLevel, set_log
from .helpers import partition
from .state_io import get_state, set_state
from .swap_system import swap_system

try:
    import pytest
except ModuleNotFoundError:
    pass
else:
    pytest.register_assert_rewrite("cosapp.utils.testing")

__all__ = [
    "LogLevel",
    "set_log",
    "partition",
    "get_state",
    "set_state",
    "swap_system",
]



import pytest

from cosapp.ports import Port, PortType
from cosapp.systems import System
from cosapp.tools.views.markdown import system_to_md, port_to_md, PortMarkdownFormatter

# <codecell> Local test classes

class DummyPort(Port):
    def setup(self):
        self.add_variable("a", 1)
        self.add_variable("b", 2)


class AnotherPort(Port):
    def setup(self):
        self.add_variable("aaaa", 1)


class DummyPortWithDesc(Port):
    def setup(self):
        self.add_variable("foo", 1, desc="Foo variable")
        self.add_variable("bar", 2, desc="Bar variable")


class System2(System):
    def setup(self):
        self.add_inward({"data1": 9.0, "data2": 11.0, "data3": 13.0})
        self.add_outward({"local1": 7.0, "a": 14.0, "b": 21.0})
        self.add_output(AnotherPort, "other")


class System3(System):
    def setup(self):
        self.add_input(DummyPort, "entry")
        self.add_output(DummyPort, "exit")


class System4(System):
    def setup(self):
        self.add_inward("X", -1.0)
        self.add_equation("X == 1.")


class DynamicSystem2(System2):
    def setup(self):
        super().setup()
        self.add_transient("A", der="a")
        self.add_rate("db_dt", source="b")


class DynamicSystem3(System3):
    def setup(self):
        super().setup()
        self.add_transient("A", der="entry.a")
        self.add_transient("B", der="entry.b")
        self.add_rate("db_out", source="exit.b")


@pytest.fixture(scope='function')
def composite():
    """Returns composite system with ports and children"""
    top = System3('top')
    top.add_child(System4('sub'))
    top.add_child(DynamicSystem3('dyn'))
    return top


# <codecell>

# TODO("be sure the test is complete")
def test_system_to_md():
    s = System("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "### Inputs" not in markdown
    assert "### Outputs" not in markdown
    assert "### Residues" not in markdown
    assert f"`{System.INWARDS}`: ExtensiblePort" not in markdown
    assert f"`{System.OUTWARDS}`: ExtensiblePort" not in markdown

    s = System2("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "### Inputs" in markdown
    assert "### Outputs" in markdown
    assert "### Residues" not in markdown
    assert f"`{System.INWARDS}`: ExtensiblePort" in markdown
    assert f"`{System.OUTWARDS}`: ExtensiblePort" in markdown

    s = System3("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "### Inputs" in markdown
    assert "### Outputs" in markdown
    assert "### Residues" not in markdown
    assert f"`{System.INWARDS}`: ExtensiblePort" not in markdown
    assert f"`{System.OUTWARDS}`: ExtensiblePort" not in markdown

    s = System4("s")
    markdown = system_to_md(s)
    assert "### Child components" not in markdown
    assert "### Inputs" in markdown
    assert "### Outputs" not in markdown
    assert "### Residues" in markdown
    assert f"`{System.INWARDS}`: ExtensiblePort" in markdown
    assert f"`{System.OUTWARDS}`: ExtensiblePort" not in markdown

    u = System("child")
    s.add_child(u)
    markdown = system_to_md(s)
    assert "### Child components" in markdown

    class NewS(System):
        tags = ["dev"]

    s = NewS("s")
    markdown = system_to_md(s)
    assert "**Tags**" in markdown


@pytest.mark.parametrize("direction", PortType)
@pytest.mark.parametrize(
    "PortCls, expected",
    [
        (DummyPort, "  **a**: 1 | &nbsp;\n  **b**: 2 | &nbsp;"),
        (DummyPortWithDesc, "  **foo**: 1 | Foo variable\n  **bar**: 2 | Bar variable"),
    ],
)
def test_port_to_md(PortCls: type, direction, expected):
    p = PortCls("p", direction)
    div_header = PortMarkdownFormatter.div_header()
    table_header = f"\n{div_header}\n\n|  |  |\n---|---\n"
    header = f"`p`: {PortCls.__name__}\n{table_header}"
    footer = "\n</div>"
    assert port_to_md(p) == f"{header}{expected}{footer}"


def test_PortMarkdownFormatter_div_header():
    assert PortMarkdownFormatter.div_header() == (
        r"<div class='cosapp-port-table' style='margin-left: 25px; margin-top: -12px'>"
        r"<style type='text/css'>"
        r".cosapp-port-table >table >thead{display: none}"
        r".cosapp-port-table tbody tr{background: rgba(255, 255, 255, 0)!important}"
        r".cosapp-port-table tbody tr:hover{background: #e1f5fe!important}"
        r".cosapp-port-table >table {margin-left: unset; margin-right: unset}"
        r"</style>"
    )


def test_PortMarkdownFormatter_wrap():
    assert PortMarkdownFormatter.wrap("foo") == [
        "",
        PortMarkdownFormatter.div_header(),
        "",
        "|  |  |",
        "---|---",
        "foo",
        "</div>",
    ]

    assert PortMarkdownFormatter.wrap(['x', 'y']) == [
        "",
        PortMarkdownFormatter.div_header(),
        "",
        "|  |  |",
        "---|---",
        "x", "y",
        "</div>",
    ]


@pytest.mark.parametrize("direction", PortType)
@pytest.mark.parametrize("PortCls, expected", [
    (
        DummyPort,
        ["`p`: DummyPort"] +
        PortMarkdownFormatter.wrap([
            "  **a**: 1 | &nbsp;",
            "  **b**: 2 | &nbsp;",
        ]),
    ),
    (
        DummyPortWithDesc,
        ["`p`: DummyPortWithDesc"] +
        PortMarkdownFormatter.wrap([
            "  **foo**: 1 | Foo variable",
            "  **bar**: 2 | Bar variable",
        ]),
    ),
    (
        DummyPortWithDesc,
        [
            "`p`: DummyPortWithDesc",
            "",
            PortMarkdownFormatter.div_header(),
            "",
            "|  |  |",
            "---|---",
            "  **foo**: 1 | Foo variable",
            "  **bar**: 2 | Bar variable",
            "</div>",
        ],
    ),
])
def test_PortMarkdownFormatter_content(PortCls: type, direction, expected):
    p = PortCls("p", direction)
    mdt = PortMarkdownFormatter(p)
    assert mdt.content() == expected


@pytest.mark.parametrize("direction", PortType)
@pytest.mark.parametrize("PortCls, expected", [
    (
        DummyPort,
        PortMarkdownFormatter.wrap([
            "  **a**: 1 | &nbsp;",
            "  **b**: 2 | &nbsp;",
        ]),
    ),
    (
        DummyPortWithDesc,
        PortMarkdownFormatter.wrap([
            "  **foo**: 1 | Foo variable",
            "  **bar**: 2 | Bar variable",
        ]),
    ),
])
def test_PortMarkdownFormatter_var_repr(PortCls: type, direction, expected):
    p = PortCls("p", direction)
    mdt = PortMarkdownFormatter(p)
    assert mdt.var_repr() == expected


def test_PortMarkdownFormatter_markdown(composite):
    top = composite
    assert top.name == "top"

    mdt = PortMarkdownFormatter(top.entry)
    assert mdt.markdown() == mdt.markdown(contextual=True)  # test default
    assert mdt.markdown(contextual=True).startswith("`top.entry`: DummyPort")
    assert mdt.markdown(contextual=False).startswith("`entry`: DummyPort")

    mdt = PortMarkdownFormatter(top.exit)
    assert mdt.markdown(contextual=True).startswith("`top.exit`: DummyPort")
    assert mdt.markdown(contextual=False).startswith("`exit`: DummyPort")

    mdt = PortMarkdownFormatter(top.sub.inwards)
    assert mdt.markdown(contextual=True).startswith("`top.sub.inwards`: ExtensiblePort")
    assert mdt.markdown(contextual=False).startswith("`inwards`: ExtensiblePort")

    mdt = PortMarkdownFormatter(top.dyn.entry)
    assert mdt.markdown(contextual=True).startswith("`top.dyn.entry`: DummyPort")
    assert mdt.markdown(contextual=False).startswith("`entry`: DummyPort")

    # Full content test
    top.dyn.entry.a = 3.14
    top.dyn.entry.b = -0.5
    expected = (
        "`top.dyn.entry`: DummyPort\n"
        f"\n{PortMarkdownFormatter.div_header()}\n"
        "\n|  |  |\n---|---\n"
        "  **a**: 3.14 | &nbsp;\n  **b**: -0.5 | &nbsp;"
        "\n</div>"
    )
    assert mdt.markdown() == expected



import pytest
import numpy as np

from cosapp.systems import System
from cosapp.recorders import DataFrameRecorder
from cosapp.drivers import RungeKutta
from typing import Tuple


class PointDynamics(System):
    """Point mass dynamics"""
    def setup(self):
        self.add_inward("mass", 1.0)
        self.add_inward("acc_ext", np.zeros(3))
        self.add_inward("force_ext", np.zeros(3))

        self.add_outward("force", np.zeros(3))
        self.add_outward("acc", np.zeros(3))

    def compute(self):
        self.force = self.force_ext + self.mass * self.acc_ext
        self.acc = self.force / self.mass


class PointFriction(System):
    """Point mass ~ v**2 friction model"""
    def setup(self):
        self.add_inward('v', np.zeros(3), desc="Velocity")
        self.add_inward('cf', 0.1, desc="Friction coefficient")

        self.add_outward("force", np.zeros(3))

    def compute(self):
        self.force = (-self.cf * np.linalg.norm(self.v)) * self.v


class PointMass(System):
    """Free fall model with friction.
    """
    def setup(self):
        self.add_child(PointFriction('friction'), pulling=['cf', 'v'])
        self.add_child(PointDynamics('dynamics'), pulling={
            'mass': 'mass',
            'force': 'force',
            'acc_ext': 'g',
            'acc': 'a',
        })

        self.connect(self.friction, self.dynamics, {"force": "force_ext"})

        self.add_transient('v', der='a')
        self.add_transient('x', der='v')

        self.g = np.r_[0, 0, -9.81]
        self.exec_order = ['friction', 'dynamics']


class BouncingBall(PointMass):
    """Bouncing ball model, combining a free fall model with a rebound event.
    """
    def setup(self):
        super().setup()
        self.add_event('rebound', trigger="x[2] <= 0")

    def transition(self):
        if self.rebound.present:
            v = self.v
            if abs(v[2]) < 1e-6:
                v[2] = 0
            else:
                v[2] *= -1


@pytest.fixture
def ball() -> BouncingBall:
    return BouncingBall('ball')


@pytest.fixture
def ball_case(ball: BouncingBall) -> Tuple[BouncingBall, RungeKutta]:
    """Bouncing ball + driver test case"""
    driver = ball.add_driver(RungeKutta(order=3))
    driver.time_interval = (0, 4)
    driver.dt = 0.01

    # Add a recorder to capture time evolution in a dataframe
    driver.add_recorder(
        DataFrameRecorder(includes=['x', 'v', 'a']),
        period=0.05,
    )
    # Define a simulation scenario
    driver.set_scenario(
        init = {
            'x': np.array([0, 0, 0]),
            'v': np.array([8, 0, 9.5]),
        },
        values = {'mass': 1.5, 'cf': 0.2},
    )
    return ball, driver


def test_BouncingBall(ball_case: Tuple[BouncingBall, RungeKutta]):
    ball, driver = ball_case

    ball.run_drivers()

    # Retrieve recorded data
    data = driver.recorder.export_data()
    x = np.asarray(data['x'].tolist())
    # Check that all positions are above ground level,
    # within numerical tolerance
    assert min(x[:, 2]) > -1e-13
    assert len(driver.recorded_events) == 3
    assert [record.time for record in driver.recorded_events] == pytest.approx(
        [1.457205, 2.536008, 3.441106]
    )


def test_BouncingBall_final(ball_case: Tuple[BouncingBall, RungeKutta]):
    """Bouncing ball case with final rebound event.
    """
    ball, driver = ball_case
    ball.rebound.final = True

    ball.run_drivers()

    # Retrieve recorded data
    data = driver.recorder.export_data()
    x = np.asarray(data['x'].tolist())
    assert min(x[:, 2]) > -1e-13
    assert ball.x == pytest.approx([6.41626970, 0, 0])
    assert len(driver.recorded_events) == 1
    assert driver.recorded_events[0].time == pytest.approx(1.457205)


def test_BouncingBall_stop(ball: BouncingBall):
    """Bouncing ball case with stop criterion based on rebound event.
    """
    driver = ball.add_driver(RungeKutta(order=3))
    driver.time_interval = (0, 4)
    driver.dt = 0.01

    # Add a recorder to capture time evolution in a dataframe
    driver.add_recorder(
        DataFrameRecorder(includes=['x', 'v', 'a']),
        period=0.05,
    )
    # Define a simulation scenario
    driver.set_scenario(
        init = {
            'x': np.array([0, 0, 0]),
            'v': np.array([8, 0, 9.5]),
        },
        values = {'mass': 1.5, 'cf': 0.2},
        stop = ball.rebound,
    )

    ball.run_drivers()

    # Retrieve recorded data
    data = driver.recorder.export_data()
    x = np.asarray(data['x'].tolist())
    assert min(x[:, 2]) > -1e-13
    assert len(driver.recorded_events) == 1
    record = driver.recorded_events[-1]
    assert len(record.events) == 2
    assert record.time == pytest.approx(1.457205)
    assert record.events[0] is ball.rebound
    assert record.events[1] is driver.scenario.stop


def test_BouncingBall_frictionless(ball: BouncingBall):
    """Bouncing ball case with stop criterion based on rebound event.
    Check analytical solution for frictionless motion.
    Order 2 Runge-Kutta solution is expected to be exact.
    """
    driver = ball.add_driver(RungeKutta(order=2))
    driver.time_interval = (0, 4)
    driver.dt = 0.01
    # Add a recorder to capture time evolution in a dataframe
    driver.add_recorder(
        DataFrameRecorder(includes=['x', 'v', 'a']),
        period=0.05,
    )
    # Initial conditions
    x0 = np.r_[0, 0, 2]
    v0 = np.r_[8, 0, 9.5]

    # Define a simulation scenario
    driver.set_scenario(
        init = {'x': np.array(x0), 'v': np.array(v0)},
        values = {'mass': 1.0, 'cf': 0.0},
        stop = ball.rebound,
    )

    ball.run_drivers()

    # Retrieve recorded data
    data = driver.recorder.export_data()
    x = np.asarray(data['x'].tolist())
    v = np.asarray(data['v'].tolist())
    g = ball.g

    # Analytical solution
    t_rebound = -v0[2] / g[2] * (1 + np.sqrt(1 - 2 * g[2] * x0[2] / v0[2]**2))
    v_exact = lambda t: v0 + g * t
    x_exact = lambda t: x0 + (v0 + 0.5 * g * t) * t

    assert min(x[:, 2]) > -1e-14
    assert len(driver.recorded_events) == 1
    record = driver.recorded_events[-1]
    assert len(record.events) == 2
    assert record.time == pytest.approx(t_rebound, rel=1e-15)
    assert record.events[0] is ball.rebound
    assert record.events[1] is driver.scenario.stop

    # Check rebound point (before transition, hence -2 index)
    assert v[-2, :] == pytest.approx(v_exact(t_rebound), rel=1e-14)
    assert x[-2, :] == pytest.approx(x_exact(t_rebound), rel=1e-14)



"""Build a VISjs view of a System."""

import os
from typing import Dict, List, Optional, Tuple, Union

from cosapp.drivers import Driver
from cosapp.ports.port import BasePort, ExtensiblePort
from cosapp.systems import System
from cosapp.tools import templates

from .baseRenderer import BaseRenderer


class VisJsRenderer(BaseRenderer):
    """
    Utility class to export a system as HTML using vis.JS library.

    Parameters
    ----------
    system : System
        System to export
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """

    __globals = None

    def __init__(self, system: System, embeddable=False):
        super().__init__(system, embeddable)

    @staticmethod
    def get_system_label(system: System):
        label = system.name
        lmax = 7
        if len(label) > lmax:
            dots = "\u2026"  # single character
            label = label[:lmax-1] + dots
        return label

    def get_system_title(self, system: System):
        head = self.system
        if system is head:
            name = head.name
        else:
            name = head.get_path_to_child(system)
        return f"{name} - {system.__class__.__name__}"

    def get_data(self, focus_system: Optional[System] = None) -> Dict:
        """Convert the `System` in a dictionary with 4 keys: title, nodes, edges and groups.

        The title is a string with the Module name.
        The nodes is a list of dictionaries defining node for vis JS library.
        The edges is a list of dictionaries defining edge for vis JS library.
        The groups is a list of string defining the nodes clusters for vis JS library.

        Parameters
        ----------
        focus_system : System or None
            System on which the visualization should focus by default

        Returns
        -------
        dict
            Dictionary containing elements to draw the `System` using visJS library
        """
        def visjs_data(system, focus_system: Optional[System] = None) -> Dict:
            if system.parent is not None:
                center = system if focus_system is None else focus_system
                return visjs_data(system.parent, center)

            def get_component2id(
                system: System,
                cmpt2id: Dict[Union[System, BasePort], int],
                node_id: int,
                edges: List[Dict],
            ) -> Tuple[Dict[Union[System, BasePort], int], int, List[Dict]]:
                """Set the mapping between the `Module` and their unique id.

                Parameters
                ----------
                system : Module
                    The parent `Module` for which id must be set
                cmpt2id : Dict[Union[Module, BasePort], int]
                    Mapping of `Module` or port and its id
                node_id : int
                    Latest used id
                edges : List[Dict]
                    List of connection edges

                Returns
                -------
                tuple (List[Dict], Dict[Union["Module", "BasePort"], int], int)
                    The first is the mapping between each involved system or port and an unique id,
                    the second is the latest used id and
                    the third index is the list of edges (one per connector).
                """

                def get_driver2id(driver, cmpt2id, node_id, edges):
                    if driver not in cmpt2id:
                        cmpt2id[driver] = node_id
                        node_id += 1

                    previous_child = None
                    for child in driver.children.values():
                        cmpt2id, node_id, edges = get_driver2id(
                            child, cmpt2id, node_id, edges
                        )
                        # Connection between same level drivers
                        if previous_child is not None: 
                            edges.append(
                                {
                                    "from": cmpt2id[previous_child],
                                    "to": cmpt2id[child],
                                    "arrows": "to",
                                    "title": "",
                                    "hidden": True,
                                    "physics": False,
                                }
                            )
                        previous_child = child
                        # Connection with parent driver
                        edges.append(
                            {
                                "from": cmpt2id[driver],
                                "to": cmpt2id[child],
                                "title": "",
                                "hidden": True,
                                "physics": False,
                            }
                        )

                    return cmpt2id, node_id, edges

                def connectors_to_visJS(
                    system: System,
                    component_id: Optional[Dict[Union[System, BasePort], int]] = None,
                    node_id: int = 1,
                ) -> Tuple[List[Dict], Dict[Union[System, BasePort], int], int]:
                    """Build the dictionary the list of edges.

                    Parameters
                    ----------
                    system : System
                        System owning the connectors
                    component_id : Dict[Union[System, Port], int], optional
                        Mapping of `System` or `Port` and its unique id; default None
                    node_id : int, optional
                        Latest used id; default 1

                    Returns
                    -------
                    tuple (List[Dict], Dict[Union["System", "BasePort"], int], int)
                        First index is the list of edges (one per connector);
                        second index is the mapping between each involved system/port and a unique id;
                        third index is the latest id used.
                    """
                    edges = list()

                    for connection in system.all_connectors():
                        supplier = connection.source.owner
                        if supplier.children:  # Insert port as node
                            supplier = connection.source

                        if supplier not in component_id:
                            component_id[supplier] = node_id
                            node_id += 1

                        target = connection.sink.owner
                        if target.children:  # Insert port as node
                            target = connection.sink

                        if target not in component_id:
                            component_id[target] = node_id
                            node_id += 1

                        edge = {
                            "from": component_id[supplier],
                            "to": component_id[target],
                            "arrows": "to",
                            "title": "",
                        }
                        if connection.is_mirror():
                            edge["title"] = type(connection.source).__name__
                        else:
                            to = "&#8594;"  # rightarrow
                            edge["title"] = ", ".join(
                                origin if origin == target else f"{origin}{to}{target}"
                                for target, origin in connection.mapping.items()
                            ).join("[]")

                        edges.append(edge)

                        # if supplier or target is top-system, set edge.length = 0
                        if (
                            connection.source.owner is connection.sink.owner.parent
                            or connection.source.owner.parent is connection.sink.owner
                        ):
                            edges[-1]["length"] = 0

                    return edges, component_id, node_id

                for driver in system.drivers.values():
                    cmpt2id, node_id, edges = get_driver2id(driver, cmpt2id, node_id, edges)

                for cpt in system.children.values():
                    if cpt not in cmpt2id:
                        cmpt2id[cpt] = node_id
                        node_id += 1

                    if cpt.children:
                        cmpt2id, node_id, edges = get_component2id(cpt, cmpt2id, node_id, edges)

                local_edges, cmpt2id, node_id = connectors_to_visJS(system, cmpt2id, node_id)
                edges.extend(local_edges)

                return cmpt2id, node_id, edges

            cmpt2id = dict()
            id = 1
            edges = list()
            cmpt2id, id_init, edges = get_component2id(system, cmpt2id, id, edges)

            nodes = list()
            # Add hidden node for the top system
            nodes.append(
                dict(
                    name=system.name,
                    label=self.get_system_label(system),
                    title=self.get_system_title(system),
                    id=id_init,
                    level=0,
                    hidden=True,
                    physics=False,
                    group=f".{system.name}",
                )
            )

            # Add edges from uppest level to its children but deactivate physics on
            # them. The goal is to have them available for the hierarchical view but
            # not playing any role in the default layout.
            for child in system.children.values():
                edges.append(
                    {
                        "from": id_init,
                        "to": cmpt2id[child],
                        "hidden": True,
                        "physics": False,
                    }
                )

            groups = list()
            for c, id in cmpt2id.items():
                if isinstance(c, BasePort):
                    ref = dict(
                        title=f"{c.full_name(trim_root=True)}",
                        id=id,
                        group=f"{c.owner.full_name()}",
                        mass=2,
                    )
                    if not isinstance(c, ExtensiblePort):
                        ref["title"] += f" - {type(c).__name__}"

                    # Add edge to bind to central owner node
                    edge = {
                        "from": id,
                        "to": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                        "hidden": True,
                    }
                    if id_init in (edge["from"], edge["to"]):
                        edge["physics"] = False
                    edges.append(edge)

                elif isinstance(c, Driver):
                    # Add driver node
                    ref = dict(
                        label=type(c).__name__,
                        title=f"{c.name} - {type(c).__name__}",
                        id=id,
                        level=-0.7,
                        shape="box",
                        hidden=True,
                        physics=False,
                        group=f"{c.owner.full_name()}",
                    )

                    # Add edge to Module owner if top driver
                    if c.parent is None:
                        edges.append(
                            {
                                "from": cmpt2id[c.owner] if c.owner in cmpt2id else id_init,
                                "to": cmpt2id[c],
                                "dashes": True,
                                "hidden": True,
                                "physics": False,
                            }
                        )
                    else:

                        def driver_depth(driver: Driver, depth: int = 0) -> int:
                            if driver.parent is None:
                                return depth
                            else:
                                return driver_depth(driver.parent, depth + 1)

                        ref["level"] += 0.2 * driver_depth(c)

                else:
                    # c is a System
                    ref = dict(
                        name=c.name,
                        label=self.get_system_label(c),
                        title=self.get_system_title(c),
                        id=id,
                        widthConstraint={'maximum': '10px'},
                    )

                    if c.children:  # Component containing component => Cluster
                        ref["hidden"] = True

                        # Add edge to force binding to central system node
                        for child in c.children.values():
                            edges.append({"from": id, "to": cmpt2id[child], "hidden": True})

                    ref["group"] = c.parent.full_name()

                group = ref["group"]
                ref.setdefault("level", 0)
                ref["level"] += group.count(".") + 1  # Set hierarchical level

                nodes.append(ref)
                if group not in groups:
                    groups.append(group)

            # Sort the groups by system to ensure that levels are clustered bottom-up
            groups.sort(reverse=True)

            data = {
                "title": self.get_system_label(system),
                "nodes": nodes,
                "edges": edges,
                "groups": groups,
            }

            if focus_system is not None:
                data["center_id"] = cmpt2id[focus_system]

            return data

        return visjs_data(self.system, focus_system)

    def html_content(self) -> str:
        """Returns vis.JS HTML content of renderer"s system as a character string."""
        elements = self.get_data()
        common = self.get_globals()
        elements["visJS"] = common["visJS"]
        elements["visCSS"] = common["visCSS"]
        html_begin_tags = common["html_begin_tags"]
        html_end_tags = common["html_end_tags"]
        template = common["template"]
        
        content = template.render(elements)
        if not self.embeddable:
            html_begin_tags = html_begin_tags.replace("{{title}}", elements["title"])
            content = html_begin_tags + content + html_end_tags

        return content

    @classmethod
    def html_resources(cls) -> Dict[str, str]:
        """
        Return a dictionary of two keys.

        - `visJS` key holds the content of vis.js library.
        - `visCSS` key holds the css style of plot.

        Returns
        -------
        Dict[str, Dict]
            The necessary resources to render Jinja template
        """
        ressource_dir = os.path.dirname(os.path.abspath(templates.__file__))
        libs_dir = os.path.join(ressource_dir, "lib")

        with open(os.path.join(libs_dir, "vis.min.js"), "rb") as f:
            vis_js = f.read().decode("utf-8")
        with open(os.path.join(libs_dir, "vis.min.css"), "rb") as f:
            vis_css = f.read().decode("utf-8")

        return {"visJS": vis_js, "visCSS": vis_css}
    
    @classmethod
    def html_template(cls) -> "Template":
        """
        Return the Jinja template used to create HTML file.

        Returns
        -------
        Template
        """
        try:
            from jinja2 import Environment, PackageLoader
        except ImportError:
            raise ImportError("jinja2 needs to be installed to export to vis JS.")

        env = Environment(
            loader=PackageLoader("cosapp.tools", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = env.get_template("system_repr.html")
        return template
    
    @classmethod
    def get_globals(cls):
        cg = cls.__globals
        if cg is None:
            cg = super().get_globals()
        return cg


def visjs_html(system: System, embeddable=False) -> str:
    """Returns vis.JS HTML content of `system`.

    Parameters
    ----------
    system : System
        System to export

    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    renderer = VisJsRenderer(system, embeddable)
    return renderer.html_content()


def to_visjs(system: System, filename: str, embeddable=False) -> None:
    """Export a system as HTML using vis.JS library.

    Parameters
    ----------
    system : System
        System to export
    filename : str
        Filename to write to
    embeddable: bool, optional
        Is the HTML to be embedded in an existing page? Default: False
    """
    ext = os.path.splitext(filename)[1]
    html_ext = f"{os.extsep}html"
    if ext != html_ext:
        filename += html_ext

    rendered_html = visjs_html(system, embeddable)
    with open(filename, mode="w", encoding="utf-8") as f:
        f.write(rendered_html)



import abc
import logging
from typing import Optional

from cosapp.drivers.runonce import RunOnce

logger = logging.getLogger(__name__)


class IterativeCase(RunOnce):
    """Abstract interface to children cases for a :py:class:`~cosapp.drivers.abstractsolver.AbstractSolver`.

    Parameters
    ----------
    name: str, optional
        Name of the `Module`
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
    **kwargs : Dict[str, Any]
        Optional keywords arguments
    """

    __slots__ = ()

    def __init__(self,
        name: str,
        owner: Optional["cosapp.systems.System"] = None,
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Driver`.
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; defaults to `None`.
        **kwargs:
            Additional keywords arguments forwarded to base class.
        """
        super().__init__(name, owner, **kwargs)
        self.reset_problem()

    def _set_owner(self, system: Optional["cosapp.systems.System"]) -> bool:
        defined = self.owner is not None
        changed = super()._set_owner(system)
        if changed:
            self.reset_problem()
            if defined:
                logger.warning(
                    f"System owner of Driver {self.name!r} has changed. Mathematical problem has been cleared."
                )

    @abc.abstractmethod
    def reset_problem(self) -> None:
        """Reset mathematical problem(s) defined on case."""
        pass

    def _postcompute(self) -> None:
        """Actions to carry out after the :py:meth:`~cosapp.drivers.runonce.RunOnce.compute` method call.

        This gathers the residues for this point and undo the variable status changes
        """
        # Request the residues of the current case to be updated
        for residue in self.get_problem().residues.values():
            residue.update()

        super()._postcompute()

    # TODO No longer needed after the suppression of RunOptim -> integrate with RunSingleCase



from numbers import Number
from typing import Optional

from cosapp.drivers.time.interfaces import ExplicitTimeDriver


class EulerExplicit(ExplicitTimeDriver):
    def __init__(
        self,
        name="Euler",
        owner: Optional["cosapp.systems.System"] = None,
        **options
    ):
        """Initialize driver

        Parameters
        ----------
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which driver belongs; defaults to `None`
        name: str, optional
            Name of the `Driver`.
        **options : Dict[str, Any]
            Optional keywords arguments; may contain time step and interval, with keys `dt` and `time_interval`
        """
        super().__init__(name, owner, **options)

    def _update_transients(self, dt: Number) -> None:
        """
        Time integration of transient variables over time step `dt` by explicit Euler scheme.
        """
        for x in self._transients.values():
            x.value += x.d_dt * dt



from __future__ import annotations
import logging
from collections import OrderedDict
from numbers import Number
from dataclasses import dataclass, field
from typing import (
    Any, Dict, Iterable, Optional,
    Sequence, Union, Tuple, List,
    Set, Callable, NamedTuple,
)

import numpy

from cosapp.core.variableref import VariableReference
from cosapp.core.numerics.boundary import Boundary, Unknown, TimeUnknown, TimeDerivative
from cosapp.core.numerics.residues import Residue, DeferredResidue
from cosapp.utils.naming import natural_varname
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


def default_array_factory(shape=0, dtype=float):
    """Default factory for dataclass fields."""
    def factory():
        return numpy.empty(shape, dtype=dtype)
    return factory


@dataclass
class SolverResults:
    """Data class for solver solution

    Attributes
    ----------
    - x [numpy.ndarray[float]]:
        Solution vector.
    - fun [numpy.ndarray[float]]:
        Values of the objective function.
    - success [bool]:
        Whether or not the solver exited successfully.
    - message [str]:
        Description of the cause of the termination.
    - tol [float, optional]:
        Tolerance level; `NaN` if not available.
    - jac [numpy.ndarray[float], optional]
        Values of function Jacobian; None if not available.
    - jac_lup [tuple(numpy.ndarray[float], numpy.ndarray[int]), optional]
        LU decomposition of Jacobian given as tuple (LU, perm); (None, None) if not available.
    - jac_calls [int]:
        Number of calls to the Jacobian matrix calculation.
    - fres_calls [int]:
        Number of calls to the residues function.
    - jac_errors [dict, optional]:
        Dictionary with singular Jacobian matrix error indexes.
    - trace [list[dict[str, Any]]]:
        Resolution history (if requested) with the evolution of x, residues and jacobian.
    """
    x: numpy.ndarray = field(default_factory=default_array_factory())
    fun: numpy.ndarray = field(default_factory=default_array_factory())
    success: bool = False
    message: str = ''
    tol: float = numpy.nan
    jac: Optional[numpy.ndarray] = None
    jac_lup: Optional[Tuple[numpy.ndarray, numpy.ndarray]] = (None, None)
    jac_calls: int = 0
    fres_calls: int = 0
    jac_errors: dict = field(default_factory=dict)
    trace: List[Dict[str, Any]] = field(default_factory=list)


class WeakDeferredResidue(NamedTuple):
    deferred: DeferredResidue
    weak: bool = False

    @property
    def target(self) -> str:
        """str: targetted quantity"""
        return self.deferred.target

    @property
    def context(self):
        """System: evaluation context of residue"""
        return self.deferred.context

    @property
    def variables(self) -> Set[str]:
        """Set[str]: names of variables involved in residue"""
        return self.deferred.variables

    def target_value(self) -> Any:
        """Evaluates and returns current value of target"""
        return self.deferred.target_value()

    def equation(self) -> str:
        """Returns target equation with updated lhs value"""
        return self.deferred.equation()

    def make_residue(self, reference=None) -> Residue:
        """Generates the residue corresponding to equation 'target == value(target)'"""
        return self.deferred.make_residue(reference)


class MathematicalProblem:
    """Container object for unknowns and equations.

    Parameters
    ----------
    name : str
        Name of the mathematical problem
    context : cosapp.systems.System
        Context in which the mathematical problem will be evaluated.
    """
    def __init__(self, name: str, context: Optional["System"]) -> None:
        from cosapp.systems import System
        check_arg(context, 'context', (System, type(None)))
        # TODO add point label to associate set of equations with Single Case
        self._name = name  # type: str
        self._context: System = context
        self._unknowns = OrderedDict()  # type: Dict[str, Unknown]
        self._residues = OrderedDict()  # type: Dict[str, Residue]
        self._transients = OrderedDict()  # type: Dict[str, TimeUnknown]
        self._rates = OrderedDict()  # type: Dict[str, TimeDerivative]
        self._targets = OrderedDict()  # type: Dict[str, WeakDeferredResidue]

    def __repr__(self) -> str:
        lines = []
        indent = "  "
        def format_unknown(items: Tuple[str, Unknown]) -> str:
            key, unknown = items
            value = unknown.default_value
            if value is None:
                value = unknown.value
            return f"{indent}{key} = {value}"

        if self.unknowns:
            lines.append("Unknowns")
            lines.extend(
                map(format_unknown, self.unknowns.items())
            )
        if self.residues or self.deferred_residues:
            lines.append("Equations")
            lines.extend(
                f"{indent}{key} := {residue.value}"
                for key, residue in self.residues.items()
            )
            lines.extend(
                f"{indent}{deferred.equation()} (target)"
                for deferred in self.deferred_residues.values()
            )
        return "\n".join(lines) if lines else "empty problem"

    @property
    def name(self) -> str:
        """str : Mathematical system name."""
        return self._name

    @property
    def context(self) -> Optional['cosapp.systems.System']:
        """cosapp.systems.System or None: Context in which mathematical objects are evaluated."""
        return self._context

    @context.setter
    def context(self, context: Optional['cosapp.systems.System']):
        if self._context is None:
            self._context = context
        elif context is not self._context:
            raise ValueError(f"Context is already set to {self._context.name!r}.")

    @property
    def residues(self) -> Dict[str, Residue]:
        """Dict[str, Residue]: Residue dictionary defined in problem."""
        return self._residues

    @property
    def unknowns(self) -> Dict[str, Unknown]:
        """Dict[str, Unknown]: Unknown dictionary defined in problem."""
        return self._unknowns

    def residue_vector(self) -> numpy.ndarray:
        """numpy.ndarray: Residue values stacked into a vector."""
        return self.__as_vector(self.residues)

    def unknown_vector(self):
        """numpy.ndarray: Unknown values stacked into a vector."""
        return self.__as_vector(self.unknowns)

    def __as_vector(self, collection: dict):
        values = tuple(
            element.value for element in collection.values()
        )
        return numpy.hstack(values) if values else numpy.empty(0)

    def residue_names(self) -> Tuple[str]:
        """Tuple[str]: Names of residues, flattened to have the same size as `residue_vector()`."""
        names = []
        for name, residue in self.residues.items():
            n_values = numpy.size(residue.value)
            if n_values > 1:
                names.extend(f"{name}[{i}]" for i in range(n_values))
            else:
                names.append(name)
        return tuple(names)

    def unknown_names(self) -> Tuple[str]:
        """Tuple[str]: Names of unknowns flatten to have the same size as `unknown_vector()`."""
        names = []
        for unknown in self.unknowns.values():
            if unknown.mask is None:
                names.append(unknown.name)
            else:
                basename = unknown.basename
                ref_size = numpy.size(unknown.ref.value)
                names.extend(
                    f"{basename}[{i}]" for i in numpy.arange(ref_size)[unknown.mask.flatten()]
                )
        return tuple(names)

    @property
    def n_unknowns(self) -> int:
        """int: Number of unknowns."""
        return sum(
            numpy.size(unknown.value) 
            for unknown in self.unknowns.values()
        )

    @property
    def n_equations(self) -> int:
        """int: Number of equations (including deferred equations)."""
        n_equations = sum(
            numpy.size(residue.value)
            for residue in self.residues.values()
        )
        n_equations += sum(
            numpy.size(deferred.target_value())
            for deferred in self.deferred_residues.values()
        )
        return n_equations

    @property
    def shape(self) -> Tuple[int, int]:
        """(int, int) : Number of unknowns and equations."""
        return (self.n_unknowns, self.n_equations)

    def is_empty(self) -> bool:
        return self.shape == (0, 0)

    def add_unknown(self,
        name: Union[str, Iterable[Union[dict, str, Unknown]]],
        max_abs_step: Number = numpy.inf,
        max_rel_step: Number = numpy.inf,
        lower_bound: Number = -numpy.inf,
        upper_bound: Number = numpy.inf,
        mask: Optional[numpy.ndarray] = None,
    ) -> MathematicalProblem:
        """Add unknown variables.

        You can set variable one by one or provide a list of dictionary to set multiple variable at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        name : str or Iterable of dictionary or str
            Name of the variable or list of variable to add
        max_rel_step : float, optional
            Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
        max_abs_step : float, optional
            Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
        lower_bound : float, optional
            Lower bound on which the solver solution is saturated; default -numpy.inf
        upper_bound : float, optional
            Upper bound on which the solver solution is saturated; default numpy.inf
        mask : numpy.ndarray or None
            Mask of unknown values in the vector variable

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_context("unknowns")
        context = self.context

        def add_unknown(
            name: str,
            max_abs_step: Number = numpy.inf,
            max_rel_step: Number = numpy.inf,
            lower_bound: Number = -numpy.inf,
            upper_bound: Number = numpy.inf,
            mask: Optional[numpy.ndarray] = None,
        ):
            """Set numerical limitations on variables.

            Parameters
            ----------
            context: cosapp.systems.System
                Object in which the unknown is defined
            name : str
                Name of the variable
            max_rel_step : float, optional
                Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
            max_abs_step : float, optional
                Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
            lower_bound : float, optional
                Lower bound on which the solver solution is saturated; default -numpy.inf
            upper_bound : float, optional
                Upper bound on which the solver solution is saturated; default numpy.inf
            mask : numpy.ndarray or None
                Mask of unknown values in the vector variable
            """
            # TODO we have a problem here if a vector variable is defined as unknown partially multiple times
            #   Example a = [1, 2, 3] with Unknown1 = a[0] & Unknown2 = a[2]

            unknown = Unknown(context, name, max_abs_step, max_rel_step, lower_bound, upper_bound, mask)
            if unknown.name in self._unknowns:
                logger.info(
                    f"Variable {name!r} is already declared as unknown in {self.name!r}."
                )
            else:
                self._unknowns[unknown.name] = unknown

        params = (max_abs_step, max_rel_step, lower_bound, upper_bound, mask)

        if isinstance(name, str):
            add_unknown(name, *params)
        else:
            for unknown in name:
                if isinstance(unknown, Unknown):
                    current_to_context = context.get_path_to_child(unknown.context)
                    new_name = f"{current_to_context}.{name}" if current_to_context else unknown.name
                    if new_name in self._unknowns:
                        logger.warning(
                            "Unknown {!r} already exists in mathematical system {!r}. "
                            "It will be overwritten.".format(new_name, self.name)
                        )
                    self._unknowns[new_name] = unknown
                elif isinstance(unknown, str):
                    add_unknown(unknown, *params)
                else:
                    add_unknown(**unknown)

        return self

    def add_equation(self,
        equation: Union[str, Iterable[Union[dict, str, Tuple[str, str]]]],
        name: Optional[str] = None,
        reference: Union[Number, numpy.ndarray, str] = 1,
    ) -> MathematicalProblem:
        """Add residue equation.

        You can add residue equation one by one or provide a list of dictionary to add multiple equation at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        equation : str or Iterable of str of the kind 'lhs == rhs'
            Equation or list of equations to add
        name : str, optional
            Name of the equation; default None => 'lhs == rhs'
        reference : Number, numpy.ndarray or "norm", optional
            Reference value(s) used to normalize the equation; default is 1.
            If value is "norm", actual reference value is estimated from order of magnitude.

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_context("equations")
        context = self.context

        def add_residue(equation, name=None, reference=1):
            """Add residue from equation."""
            residue = Residue(context, equation, name, reference)
            self._residues[residue.name] = residue

        if isinstance(equation, str):
            add_residue(equation, name, reference)
        else:
            for eq in equation:
                if isinstance(eq, str):
                    add_residue(eq)
                elif isinstance(eq, dict):
                    add_residue(**eq)
                else:
                    add_residue(*eq)

        return self

    def add_target(self,
        expression: Union[str, Iterable[str]],
        reference: Union[Number, numpy.ndarray, str] = 1,
        weak = False,
    ) -> MathematicalProblem:
        """Add deferred equation.

        Parameters
        ----------
        expression: str
            Targetted expression
        reference : Number, numpy.ndarray or "norm", optional
            Reference value(s) used to normalize the (deferred) equation; default is 1.
            If value is "norm", actual reference value is estimated from order of magnitude.
        weak: bool, optional
            If True, the target is disregarded if the corresponding variable is connected; default is `False`.

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_context("targets")
        context = self.context

        def register(name, reference=1):
            deferred = DeferredResidue(context, name, reference)
            if len(deferred.variables) > 1:
                raise NotImplementedError(
                    f"Targets are only supported for single variables; got {deferred.variables}"
                )
            key = self.target_key(deferred.target)
            self._targets[key] = WeakDeferredResidue(deferred, weak)

        if isinstance(expression, str):
            register(expression, reference)
        else:
            for target in expression:
                register(target)

        return self

    def get_target_equations(self) -> List[str]:
        return [deferred.equation() for deferred in self._targets.values()]

    def get_target_residues(self) -> Dict[str, Residue]:
        return dict(
            (key, deferred.make_residue())
            for key, deferred in self._targets.items()
        )

    def activate_targets(self) -> None:
        """Activate deferred residues (targets) and incorporate them
        in the mathematical problem residue list.
        Warning: This operation is irreversible, as targets are purged.
        """
        targets = self._targets
        residues = self._residues
        for key in list(targets):
            deferred = targets.pop(key)
            residue = deferred.make_residue()
            new_key = self.target_key(residue.name)
            residues[new_key] = residue

    @staticmethod
    def target_key(target: str) -> str:
        """Returns dict key to be used for targetted quantity `target`"""
        return f"{target} (target)"

    @property
    def deferred_residues(self) -> Dict[str, WeakDeferredResidue]:
        """Dict[str, WeakDeferredResidue]: Dict of deferred residues defined for this system."""
        return self._targets

    @property
    def transients(self) -> Dict[str, TimeUnknown]:
        """Dict[str, TimeUnknown] : Unknown time-dependent numerical features defined for this system."""
        return self._transients

    def add_transient(self,
        name: str,
        der: Any,
        max_time_step: Union[Number, str] = numpy.inf,
        max_abs_step: Union[Number, str] = numpy.inf,
        pulled_from: Optional[VariableReference] = None,
    ) -> MathematicalProblem:
        """Add a time-dependent unknown.

        Parameters
        ----------
        name : str
            Name of the new time-dependent (transient) variable
        der : Any
            Name of the variable or evaluable expression defining the time derivative of transient variable
        max_time_step : Number or evaluable expression (str), optional
            Maximal time step for the time integration of the transient variable; numpy.inf by default.
        max_abs_step : Number or evaluable expression compatible with transient variable, optional
            Maximum variable step admissible over one time step; numpy.inf by default.
        pulled_from : VariableReference, optional
            Original time unknown before pulling; None by default.

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_context("transient unknowns")

        if name in self._transients:
            raise ArithmeticError(f"Variable {name!r} is already defined as a time-dependent unknown of {self.name!r}.")

        self._transients[name] = TimeUnknown(
            self.context,
            name,
            der,
            max_time_step,
            max_abs_step,
            pulled_from=pulled_from,
        )
        return self

    @property
    def rates(self) -> Dict[str, TimeDerivative]:
        """Dict[str, TimeDerivative] : Time derivatives computed during system evolution."""
        return self._rates

    def add_rate(self, name: str, source: Any, initial_value: Any = None) -> MathematicalProblem:
        """Add a time derivative.

        Parameters
        ----------
        name : str
            Name of the new time-dependent (transient) variable
        source : Any
            Name of the variable or evaluable expression whose time derivative will be computed

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.__check_context("rates")

        if name in self._rates:
            raise ArithmeticError(f"Variable {name!r} is already defined as a time-dependent unknown of {self.name!r}.")

        self._rates[name] = TimeDerivative(self.context, name, source, initial_value)
        return self

    def __check_context(self, name: str):
        if self.context is None:
            raise AttributeError(f"Owner System is required to define {name}.")

    def extend(self,
        other: MathematicalProblem,
        copy = True,
        unknowns = True,
        equations = True,
        overwrite = False,
        unknown_wrapper: Optional[Callable[[str], str]] = None,
        residue_wrapper: Optional[Callable[[str], str]] = None,
    ) -> MathematicalProblem:
        """Extend the current mathematical system with the other one.

        Parameters
        ----------
        - other [MathematicalProblem]:
            The other mathematical system to add
        - copy [bool, optional]:
            Should the objects be copied; default is `True`.
        - unknowns [bool, optional]:
            If `False`, unknowns are discarded; default is `True`.
        - equations [bool, optional]:
            If `False`, equations are discarded; default is `True`.
        - overwrite [bool, optional]:
            If `False` (default), common unknowns/equations raise `ValueError`.
            If `True`, attributes are silently overwritten.

        Returns
        -------
        MathematicalProblem
            The resulting mathematical system
        """
        no_wrapping = (unknown_wrapper is residue_wrapper is None)
        if other is self and not copy and no_wrapping:
            return self  # quick return

        if unknown_wrapper is None:
            unknown_wrapper = lambda key: key
        if residue_wrapper is None:
            residue_wrapper = lambda key: key

        from cosapp.systems import System
        context: System = self.context

        sys_paths: Dict[int, str] = dict()

        def get_path(system: System) -> str:
            nonlocal sys_paths
            key = id(system)
            try:
                path = sys_paths[key]
            except KeyError:
                sys_paths[key] = path = context.get_path_to_child(system)
            return path

        def var_key_format(path: str, varname: str):
            return f"{path}.{varname}" if path else varname

        def res_key_format(path: str, eqname: str):
            return f"{path}: {eqname}" if path else eqname

        def make_key(obj: Union[Boundary, Residue], key_format, wrapper) -> str:
            """Generic key formatter for unknowns and residues."""
            path = get_path(obj.context)
            key = key_format(path, obj.name)
            return wrapper(key)

        def variable_key(variable: Boundary) -> str:
            """Generate dict key from `variable` context."""
            return make_key(variable, var_key_format, unknown_wrapper)

        def residue_key(residue: Residue) -> str:
            """Generate dict key from `residue` context."""
            return make_key(residue, res_key_format, residue_wrapper)

        get = (lambda obj: obj.copy()) if copy else (lambda obj: obj)

        def transfer_unknowns(kind: str):
            """Transfer unknowns from other mathematical problem"""
            source = getattr(other, kind)
            destination = getattr(self, kind)
            for unknown in source.values():
                key = variable_key(unknown)
                if not overwrite and key in destination:
                    raise ValueError(f"{key!r} already exists in {self.name!r}.")
                destination[key] = get(unknown)

        if unknowns:
            transfer_unknowns('unknowns')
            transfer_unknowns('transients')
            transfer_unknowns('rates')

        if equations:
            residues = self._residues
            for name, residue in other.residues.items():
                key = residue_key(residue)
                if not overwrite and key in residues:
                    raise ValueError(f"{key!r} already exists in {self.name!r}.")
                residues[key] = get(residue)

            connectors = list(self.context.incoming_connectors())
            name2variable = other.context.name2variable
            path = get_path(other.context)

            for deferred in other._targets.values():
                targetted = list(deferred.variables)[0]
                name = deferred.target.replace(targetted, var_key_format(path, targetted))  # default
                ref = name2variable[targetted]
                port, varname = ref.mapping, ref.key
                # Check if targetted variable is a pulled output
                if port.is_output:
                    for connector in connectors:
                        pulled = (
                            connector.source is port
                            and varname in connector.source_variables()
                        )
                        if pulled:
                            alias_name = natural_varname(
                                f"{connector.sink.name}.{connector.sink_variable(varname)}"
                            )
                            original = name
                            if deferred.target == targetted:
                                name = alias_name
                            else:
                                # target is an expression involving `targetted`
                                name = name.replace(var_key_format(path, targetted), alias_name)
                            logger.info(
                                f"Target on {original!r} will be based on {name!r} in the context of {self.context.full_name()!r}"
                            )
                            break
                self.add_target(name, weak=deferred.weak)

        return self

    def clear(self) -> None:
        """Clear all mathematical elements in this problem."""
        self._unknowns.clear()
        self._residues.clear()
        self._transients.clear()
        self._rates.clear()
        self._targets.clear()

    def copy(self, activate_targets=False) -> MathematicalProblem:
        """Copy the `MathematicalSystem` object.

        Returns
        -------
        MathematicalProblem
            The duplicated mathematical problem.
        """
        new = MathematicalProblem(self.name, self.context)
        new.extend(self, copy=True)
        if activate_targets:
            new.activate_targets()
        return new

    def to_dict(self) -> Dict[str, Any]:
        """Returns a JSONable representation of the mathematical problem.
        
        Returns
        -------
        Dict[str, Any]
            JSONable representation
        """
        return {
            "unknowns": dict((name, unknown.to_dict()) for name, unknown in self.unknowns.items()),
            "equations": dict((name, equation.to_dict()) for name, equation in self.residues.items()),
            "transients": dict((name, transient.to_dict()) for name, transient in self.transients.items()),
            "rates": dict((name, rate.to_dict()) for name, rate in self.rates.items())
        }

    def validate(self) -> None:
        """Verifies that there are as much unknowns as equations defined.

        Raises
        ------
        ArithmeticError
            If the mathematical system is not closed.
        """
        n_unknowns, n_equations = self.shape
        if n_unknowns != n_equations:
            msg = "Nonlinear problem {} error: Mismatch between numbers of params [{}] and residues [{}]".format(
                self.name, n_unknowns, n_equations
            )
            logger.error(msg)
            logger.error(f"Residues: {list(self.residues) + list(self.deferred_residues)}")
            logger.error(f"Variables: {list(self.unknowns)}")
            raise ArithmeticError(msg)
